#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_safety
----------------------------------

Tests for `safety` module.
"""


import json
import os
import unittest
from collections import defaultdict
from http import HTTPStatus
from io import StringIO
from json import JSONDecodeError
from unittest.mock import Mock, patch

import click as click
from packaging.specifiers import SpecifierSet
from packaging.version import parse
from requests.exceptions import RequestException

from safety.auth import build_client_session
from safety.constants import DB_CACHE_FILE
from safety.errors import (
    DatabaseFetchError,
    DatabaseFileNotFoundError,
    InvalidCredentialError,
    MalformedDatabase,
    TooManyRequestsError,
)
from safety.formatter import SafetyFormatter
from safety.models import CVE, Package, SafetyRequirement
from safety.safety import (
    calculate_remediations,
    check,
    compute_sec_ver,
    get_announcements,
    get_closest_ver,
    get_licenses,
    ignore_vuln_if_needed,
    precompute_remediations,
    read_vulnerabilities,
    review,
)
from safety.util import get_packages_licenses, read_requirements
from tests.resources import REMEDIATIONS, SCANNED_PACKAGES, VALID_REPORT, VULNS
from tests.test_cli import get_vulnerability


class TestSafety(unittest.TestCase):

    def setUp(self) -> None:
        self.session, _ = build_client_session()
        self.maxDiff = None
        self.dirname = os.path.dirname(__file__)
        self.report = VALID_REPORT
        self.report_vulns = VULNS
        self.report_packages = SCANNED_PACKAGES
        self.report_remediations = REMEDIATIONS
        self.default_pkg = Package(**{'name': 'default_pkg', 'version': '1.22.0',
                                      'requirements': [SafetyRequirement('default_pkg==1.22.0')],
                                      'found': '/site-packages/default_pkg',
                                      'insecure_versions': ['1.22.2', '1.22.1', '1.22.0', '1.22.0rc3', '1.21.5'],
                                      'secure_versions': ['1.22.3'],
                                      'latest_version_without_known_vulnerabilities': '2.2',
                                      'latest_version': '2.2', 'more_info_url': 'https://pyup.io/package/default_pkg'})

    def test_check_from_file(self):
        reqs = StringIO("Django==1.8.1")
        packages = read_requirements(reqs)

        vulns, _ = check(
            session=self.session,
            packages=packages,
            db_mirror=os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "test_db"
            ),
            cached=0,
            ignore_vulns={},
            ignore_severity_rules=None,
            proxy={},
            telemetry=False
        )
        self.assertEqual(len(vulns), 2)

    def test_check_ignores(self):
        reqs = StringIO("Django==1.8.1")
        packages = read_requirements(reqs)

        # Second that ignore works
        ignored_vulns = {'some id': {'expires': None, 'reason': ''}}

        vulns, _ = check(
            session=self.session,
            packages=packages,
            db_mirror=os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "test_db"
            ),
            cached=0,
            ignore_vulns=ignored_vulns,
            ignore_severity_rules=None,
            proxy={},
            telemetry=False
        )
        self.assertEqual(len(vulns), 1)
        self.assertEqual(vulns[0].vulnerability_id, "some other id")

    def test_check_from_file_with_hash_pins(self):
        reqs = StringIO(("Django==1.8.1 "
                         "--hash=sha256:c6c7e7a961e2847d050d214ca96dc3167bb5f2b25cd5c6cb2eea96e1717f4ade"))
        packages = read_requirements(reqs)

        vulns, _ = check(
            session=self.session,
            packages=packages,
            db_mirror=os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "test_db"
            ),
            cached=0,
            ignore_vulns={},
            ignore_severity_rules=None,
            proxy={},
            telemetry=False
        )
        self.assertEqual(len(vulns), 2)

    def test_multiple_versions(self):
        # Probably used for external tools using check directly
        reqs = StringIO("Django==1.8.1\n\rDjango==1.7.0")
        packages = read_requirements(reqs)

        vulns, _ = check(
            session=self.session,
            packages=packages,
            db_mirror=os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "test_db"
            ),
            cached=0,
            ignore_vulns={},
            ignore_severity_rules=None,
            proxy={},
            telemetry=False
        )
        self.assertEqual(len(vulns), 4)

    def test_check_live(self):
        reqs = StringIO("insecure-package==0.1")
        packages = read_requirements(reqs)

        vulns, _ = check(
            session=self.session,
            packages=packages,
            db_mirror=False,
            cached=0,
            ignore_vulns={},
            ignore_severity_rules=None,
            proxy={},
            telemetry=False
        )

        self.assertEqual(len(vulns), 1)

    def test_check_live_cached(self):
        from safety.constants import DB_CACHE_FILE

        # Ensure the cache directory and file exist
        os.makedirs(os.path.dirname(DB_CACHE_FILE), exist_ok=True)

        # lets clear the cache first
        try:
            with open(DB_CACHE_FILE, 'w') as f:
                f.write(json.dumps({}))
        except Exception:
            pass

        reqs = StringIO("insecure-package==0.1")
        packages = read_requirements(reqs)

        vulns, _ = check(
            session=self.session,
            packages=packages,
            db_mirror=False,
            cached=60 * 60,
            ignore_vulns={},
            ignore_severity_rules=None,
            proxy={},
            telemetry=False
        )
        self.assertEqual(len(vulns), 1)

        reqs = StringIO("insecure-package==0.1")
        packages = read_requirements(reqs)
        # make a second call to use the cache
        vulns, _ = check(
            session=self.session,
            packages=packages,
            db_mirror=False,
            cached=60 * 60,
            ignore_vulns={},
            ignore_severity_rules=None,
            proxy={},
            telemetry=False
        )
        self.assertEqual(len(vulns), 1)

    def test_get_packages_licenses(self):
        reqs = StringIO("Django==1.8.1\n\rinvalid==1.0.0")
        packages = read_requirements(reqs)

        licenses_db = get_licenses(
            session=self.session,
            db_mirror=os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "test_db"
            ),
            cached=0,
            telemetry=False
        )
        self.assertIn("licenses", licenses_db)
        self.assertIn("packages", licenses_db)
        self.assertIn("BSD-3-Clause", licenses_db['licenses'])
        self.assertIn("django", licenses_db['packages'])

        pkg_licenses = get_packages_licenses(packages=packages, licenses_db=licenses_db)

        self.assertIsInstance(pkg_licenses, list)
        for pkg_license in pkg_licenses:
            license = pkg_license['license']
            version = pkg_license['version']
            if pkg_license['package'] == 'django':
                self.assertEqual(license, 'BSD-3-Clause')
                self.assertEqual(version, '1.8.1')
            elif pkg_license['package'] == 'invalid':
                self.assertEqual(license, 'unknown')
                self.assertEqual(version, '1.0.0')
            else:
                raise AssertionError(
                    "unexpected package '" + pkg_license['package'] + "' was found"
                )

    def test_get_packages_licenses_without_api_key(self):
        # without providing an API-KEY
        with self.assertRaises(InvalidCredentialError) as error:
            get_licenses(
                session=self.session,
                db_mirror=False,
                cached=0,
                telemetry=False
            )
        db_generic_exception = error.exception
        self.assertEqual(str(db_generic_exception), 'Your authentication credential is invalid. See https://docs.safetycli.com/safety-docs/support/invalid-api-key-error.')


    def test_get_packages_licenses_with_invalid_api_key(self):
        session = Mock()
        api_key = "INVALID"
        session.api_key = api_key
        session.headers = {'X-Api-Key': api_key}
        mock = Mock()
        mock.status_code = 403
        session.get.return_value = mock

        # proving an invalid API-KEY
        with self.assertRaises(InvalidCredentialError):
            get_licenses(
                session=session,
                db_mirror=False,
                cached=0,
                telemetry=False
            )

    def test_get_packages_licenses_db_fetch_error(self):
        session = Mock()
        api_key = "MY-VALID-KEY"
        session.api_key = api_key
        session.headers = {'X-Api-Key': api_key}
        mock = Mock()
        mock.status_code = 500
        session.get.return_value = mock

        with self.assertRaises(DatabaseFetchError):
            get_licenses(
                session=session,
                db_mirror=False,
                cached=0,
                telemetry=False
            )

    def test_get_packages_licenses_with_invalid_db_file(self):

        with self.assertRaises(DatabaseFileNotFoundError):
            get_licenses(
                session=self.session,
                db_mirror='/my/invalid/path',
                cached=0,
                telemetry=False
            )

    def test_get_packages_licenses_very_often(self):
        # if the request is made too often, an 429 error is raise by PyUp.io
        session = Mock()
        api_key = "MY-VALID-KEY"
        session.api_key = api_key
        session.headers = {'X-Api-Key': api_key}
        mock = Mock()
        mock.status_code = 429
        session.get.return_value = mock

        with self.assertRaises(TooManyRequestsError):
            get_licenses(
                session=session,
                db_mirror=False,
                cached=0,
                telemetry=False
            )

    def test_get_cached_packages_licenses(self):
        import copy

        licenses_db = {
            "licenses": {
                "BSD-3-Clause": 2
            },
            "packages": {
                "django": [
                    {
                        "start_version": "0.0",
                        "license_id": 2
                    }
                ]
            }
        }
        original_db = copy.deepcopy(licenses_db)

        session = Mock()
        api_key = "MY-VALID-KEY"
        session.api_key = api_key
        session.headers = {'X-Api-Key': api_key}
        mock = Mock()
        mock.status_code = 200
        mock.json.return_value = licenses_db
        session.get.return_value = mock

        # lets clear the cache first
        try:
            with open(DB_CACHE_FILE, 'w') as f:
                f.write(json.dumps({}))
        except Exception:
            pass

        # In order to cache the db (and get), we must set cached as True
        response = get_licenses(
            session=session,
            db_mirror=False,
            cached=60 * 60,  # Cached for one hour
            telemetry=False
        )
        self.assertEqual(response, licenses_db)

        # now we should have the db in cache
        # changing the "live" db to test if we are getting the cached db
        licenses_db['licenses']['BSD-3-Clause'] = 123

        resp = get_licenses(
            session=session,
            db_mirror=False,
            cached=60 * 60,  # Cached for one hour
            telemetry=False
        )

        self.assertNotEqual(resp, licenses_db)
        self.assertEqual(resp, original_db)

    def test_report_licenses_bare(self):
        reqs = StringIO("Django==1.8.1\n\rinexistent==1.0.0")
        packages = read_requirements(reqs)

        # Using DB: test.test_db.licenses.json
        licenses_db = get_licenses(
            session=self.session,
            db_mirror=os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "test_db"
            ),
            cached=0,
            telemetry=False
        )

        pkgs_licenses = get_packages_licenses(packages=packages, licenses_db=licenses_db)
        output_report = SafetyFormatter(output='bare').render_licenses([], pkgs_licenses)

        self.assertEqual(output_report, "BSD-3-Clause unknown")

    @patch('safety.formatters.json.get_report_brief_info')
    def test_report_licenses_json(self, get_report_brief_info):
        get_report_brief_info.return_value = {'scan_target': 'environment',
                                              'scanned': ['/usr/local/lib/python3.9/site-packages'],
                                              'api_key': True,
                                              'packages_found': 2,
                                              'timestamp': '2022-03-03 16:31:30',
                                              'safety_version': '2.0.0.dev6'}

        reqs = StringIO("Django==1.8.1\n\rinexistent==1.0.0")
        packages = read_requirements(reqs)

        # Using DB: test.test_db.licenses.json
        licenses_db = get_licenses(
            session=self.session,
            db_mirror=os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "test_db"
            ),
            cached=0,
            telemetry=False
        )

        pkgs_licenses = get_packages_licenses(packages=packages, licenses_db=licenses_db)
        output_report = SafetyFormatter(output='json').render_licenses([], pkgs_licenses)

        expected_result = json.dumps(
            {
                "report_meta": {
                    "scan_target": "environment",
                    "scanned": ["/usr/local/lib/python3.9/site-packages"],
                    "api_key": True,
                    "packages_found": 2,
                    "timestamp": "2022-03-03 16:31:30",
                    "safety_version": "2.0.0.dev6"},
                "announcements": [],
                "licenses": [
                    {
                        "package": "django",
                        "version": "1.8.1",
                        "license": "BSD-3-Clause"
                    },
                    {
                        "package": "inexistent",
                        "version": "1.0.0",
                        "license": "unknown",
                    }
                ]
            },
            indent=4
        )
        # Packages without license are reported as "N/A"
        self.assertEqual(output_report.rstrip(), expected_result)

    @patch("safety.util.get_used_options")
    @patch.object(click, 'get_current_context', Mock(command=Mock(name=Mock(return_value='check'))))
    def test_get_announcements_catch_request_exceptions(self, get_used_options):
        get_used_options.return_value = {'key': {'--key': 1}, 'output': {'--output': 1}}
        session = Mock()
        api_key = "somekey"
        session.api_key = api_key
        session.headers = {'X-Api-Key': api_key}
        session.get.side_effect = RequestException()
        self.assertEqual(get_announcements(session), [])

    @patch("safety.util.get_used_options")
    @patch.object(click, 'get_current_context', Mock(command=Mock(name=Mock(return_value='check'))))
    def test_get_announcements_catch_unhandled_http_codes(self, get_used_options):
        get_used_options.return_value = {'key': {'--key': 1}, 'output': {'--output': 1}}

        unhandled_status = [status for status in HTTPStatus if status != HTTPStatus.OK]

        for http_status in unhandled_status:
            mock = Mock()
            mock.status_code = http_status.value
            session = Mock()
            api_key = "somekey"
            session.api_key = api_key
            session.headers = {'X-Api-Key': api_key}
            session.get.return_value = mock

            self.assertEqual(get_announcements(session), [])

    @patch("safety.util.get_used_options")
    @patch.object(click, 'get_current_context', Mock(command=Mock(name=Mock(return_value='check'))))
    def test_get_announcements_http_ok(self, get_used_options):
        get_used_options.return_value = {}

        announcements = {
            "announcements": [{
                "type": "notice",
                "message": "You are using an outdated version of  Please upgrade to Safety version 1.2.3"
                },
                {
                "type": "error",
                "message": "You are using an vulnerable version of  Please upgrade now"
                }]
        }

        expected = announcements.get('announcements')

        mock = Mock()
        mock.status_code = HTTPStatus.OK.value
        mock.json.return_value = announcements
        session = Mock()
        api_key = "somekey"
        session.api_key = api_key
        session.headers = {'X-Api-Key': api_key}
        session.post.return_value = mock

        self.assertEqual(get_announcements(session), expected)

    @patch("safety.util.get_used_options")
    @patch.object(click, 'get_current_context', Mock(command=Mock(name=Mock(return_value='check'))))
    def test_get_announcements_wrong_json_response_handling(self, get_used_options):
        get_used_options.return_value = {}

        # wrong JSON structure
        announcements = {
                "type": "notice",
                "message": "You are using an outdated version of  Please upgrade to Safety version 1.2.3"
        }

        mock = Mock()
        mock.status_code = HTTPStatus.OK.value
        mock.json.return_value = announcements

        session = Mock()
        api_key = "somekey"
        session.api_key = api_key
        session.headers = {'X-Api-Key': api_key}
        session.get.return_value = mock

        self.assertEqual(get_announcements(session), [])

        # JSONDecodeError

        mock = Mock()
        mock.status_code = HTTPStatus.OK.value
        mock.json.side_effect = JSONDecodeError(msg='Expecting value', doc='', pos=0)
        session = Mock()
        api_key = "somekey"
        session.api_key = api_key
        session.headers = {'X-Api-Key': api_key}
        session.get.return_value = mock

        self.assertEqual(get_announcements(session), [])

    def test_ignore_vulns_by_unknown_severity(self):
        cve_no_cvss = CVE(name='PYUP-123', cvssv2=None, cvssv3=None)
        ignore_vulns = {}
        ignore_rules = {'ignore-cvss-unknown-severity': True}

        ignore_vuln_if_needed(pkg=self.default_pkg, vuln_id='1234', cve=cve_no_cvss, ignore_vulns=ignore_vulns,
                              ignore_severity_rules=ignore_rules, req=SafetyRequirement('django==2.2'))
        EXPECTED = {
            '1234': {'reason': 'Unknown CVSS severity, ignored by severity rule in policy file.', 'expires': None}}

        self.assertEqual(ignore_vulns, EXPECTED)

    def test_dont_ignore_vulns_by_unknown_severity(self):
        cve_no_cvss = CVE(name='PYUP-123', cvssv2=None, cvssv3=None)
        ignore_vulns = {}
        ignore_rules = {'ignore-cvss-unknown-severity': False}
        ignore_vuln_if_needed(pkg=self.default_pkg, vuln_id='1234', cve=cve_no_cvss, ignore_vulns=ignore_vulns,
                              ignore_severity_rules=ignore_rules, req=SafetyRequirement('django==2.2'))

        self.assertEqual(ignore_vulns, {})

    def test_ignore_vulns_by_base_score(self):
        req = SafetyRequirement('django==2.2')
        cve_cvss_medium = CVE(name='CVE-123', cvssv2={'base_score': '4,3', 'impact_score': '2,9'},
                              cvssv3={'base_score': '6.1', 'impact_score': '2.7', 'base_severity': 'MEDIUM'})
        ignore_vulns = {}
        ignore_rules = {'ignore-cvss-severity-below': 7}
        ignore_vuln_if_needed(pkg=self.default_pkg, vuln_id='1234', cve=cve_cvss_medium, ignore_vulns=ignore_vulns,
                              ignore_severity_rules=ignore_rules, req=req)
        m_b_score = cve_cvss_medium.cvssv3.get('base_score')

        EXPECTED = {
            '1234': {
                'reason': f"Ignored by severity rule in policy file, {float(m_b_score)} < 7.0", 'expires': None}
        }
        self.assertEqual(ignore_vulns, EXPECTED)

        cve_cvss_critical = CVE(name='PVE-124', cvssv2=None,
                                cvssv3={'base_score': '9.8', 'impact_score': '5.9', 'base_severity': 'CRITICAL'})
        ignore_vulns = {}
        ignore_rules = {'ignore-cvss-severity-below': 7}
        ignore_vuln_if_needed(pkg=self.default_pkg, vuln_id='1235', cve=cve_cvss_critical, ignore_vulns=ignore_vulns,
                              ignore_severity_rules=ignore_rules, req=req)
        self.assertEqual(ignore_vulns, {})

        cve_cvss_high = CVE(name='PVE-125', cvssv2=None,
                            cvssv3={'base_score': '7.0', 'impact_score': '5.9', 'base_severity': 'HIGH'})
        ignore_vulns = {}
        ignore_rules = {'ignore-cvss-severity-below': 7}
        ignore_vuln_if_needed(pkg=self.default_pkg, vuln_id='1236', cve=cve_cvss_high, ignore_vulns=ignore_vulns,
                              ignore_severity_rules=ignore_rules, req=req)
        self.assertEqual(ignore_vulns, {})

    def test_get_closest_ver(self):
        versions = get_closest_ver(versions=['1.2', '1.3', '1.3.1', '1.3.2.dev5'], version='1.3.1',
                                   spec=SpecifierSet('==1.3.1'))
        EXPECTED = {'lower': parse('1.3'), 'upper': parse('1.3.2.dev5')}
        self.assertEqual(versions, EXPECTED)

        versions = get_closest_ver(versions=['1.2', '1.3', '1.3.1'], version='1.3.1', spec=SpecifierSet('==1.3.1'))
        EXPECTED = {'lower': parse('1.3'), 'upper': None}
        self.assertEqual(versions, EXPECTED)

        versions = get_closest_ver(versions=['1.2', '1.3', '1.3.1'], version='1.2', spec=SpecifierSet('==1.2'))
        EXPECTED = {'lower': None, 'upper': parse('1.3')}
        self.assertEqual(versions, EXPECTED)

        versions = get_closest_ver(versions=[], version='1.2', spec=SpecifierSet('==1.2'))
        EXPECTED = {'lower': None, 'upper': None}
        self.assertEqual(versions, EXPECTED)

        versions = get_closest_ver(versions=['1.2', '1.3'], version=None, spec=None)
        EXPECTED = {'lower': None, 'upper': None}
        self.assertEqual(versions, EXPECTED)

    def test_precompute_remediations(self):
        numpy_pkg = {'name': 'numpy', 'version': '1.22.0', 'requirements': [SafetyRequirement('numpy==1.22.0')],
                     'found': '/site-packages/numpy',
                     'insecure_versions': ['1.22.2', '1.22.1', '1.22.0', '1.22.0rc3', '1.21.5'],
                     'secure_versions': ['1.22.3'], 'latest_version_without_known_vulnerabilities': '2.2',
                     'latest_version': '2.2', 'more_info_url': 'https://pyup.io/package/numpy'}

        vulns = [
            get_vulnerability(pkg_kwargs=numpy_pkg, vuln_kwargs={'affected_versions': ['1.22.0', '1.21.5']}),
            get_vulnerability(pkg_kwargs=numpy_pkg,
                              vuln_kwargs={'affected_versions': ['1.22.0', '1.22.0rc3']}),
            get_vulnerability(vuln_kwargs={'ignored': True}, pkg_kwargs={'name': 'django'})
        ]
        remediations = defaultdict(dict)
        package_meta = {}
        precompute_remediations(remediations=remediations, packages=package_meta, vulns=vulns,
                                secure_vulns_by_user=set())

        EXPECTED = {'numpy': Package(**numpy_pkg)}
        self.assertEqual(package_meta, EXPECTED)

        EXPECTED = {'numpy': {'==1.22.0':
                                  {'vulnerabilities_found': 2, 'version': '1.22.0',
                                   'requirement': SafetyRequirement('numpy==1.22.0'),
                                   'more_info_url': 'https://pyup.io/package/numpy'}}
                    }

        self.assertEqual(remediations, EXPECTED)

    @patch("safety.safety.is_using_api_key")
    def test_compute_sec_ver(self, is_using_api_key):
        is_using_api_key.return_value = True

        import copy

        test_filename = os.path.join(self.dirname, "test_db/insecure_full_affected_versions.json")
        db_full = None
        with open(test_filename) as f:
            db_full = json.loads(f.read())
        self.assertIsNotNone(db_full)

        npy = Package(insecure_versions=['1.22.2', '1.22.1', '1.22.0', '1.22.0rc3', '1.21.5'],
                      secure_versions=['1.22.3'], version='1.22.0', name='numpy',
                      requirements=[SafetyRequirement('numpy==1.22.0')])

        pre_pkg_meta = {'numpy': copy.copy(npy)}
        # The vuln affecting '1.21.5' was ignored by the user
        ignored_vulns = set()
        ignored_vulns.add('29')
        rem = {'numpy': {'==1.22.0': {'vulnerabilities_found': 1, 'version': '1.22.0',
                                      'requirement': SafetyRequirement('numpy==1.22.0'),
                                      'more_info_url': 'https://pyup.io/package/foo'}}}
        compute_sec_ver(remediations=rem, packages=pre_pkg_meta, secure_vulns_by_user=ignored_vulns, db_full=db_full)
        EXPECTED = {'numpy': {
            '==1.22.0': {'vulnerabilities_found': 1, 'version': '1.22.0',
                         'requirement': SafetyRequirement('numpy==1.22.0'),
                         'other_recommended_versions': ['1.21.5'],
                         'closest_secure_version': {'upper': parse('1.22.3'), 'lower': parse('1.21.5')},
                         'more_info_url': 'https://pyup.io/package/foo?from=1.22.0&to=1.22.3',
                         'recommended_version': parse('1.22.3')}}}
        self.assertEqual(rem, EXPECTED)

        pre_pkg_meta = {'numpy': copy.copy(npy)}
        ignored_vulns = set()
        rem = {'numpy': {'==1.22.0' : {'vulnerabilities_found': 2, 'version': '1.22.0',
                         'requirement': SafetyRequirement('numpy==1.22.0'),
                         'more_info_url': 'https://pyup.io/package/foo'}}}
        compute_sec_ver(remediations=rem, packages=pre_pkg_meta, secure_vulns_by_user=ignored_vulns, db_full=db_full)
        EXPECTED = {'numpy': {'==1.22.0': {'vulnerabilities_found': 2, 'version': '1.22.0',
                                           'requirement': SafetyRequirement('numpy==1.22.0'),
                                           'other_recommended_versions': [],
                                           'closest_secure_version': {'upper': parse('1.22.3'), 'lower': None},
                                           'more_info_url': 'https://pyup.io/package/foo?from=1.22.0&to=1.22.3',
                                           'recommended_version': parse('1.22.3')}}}

        self.assertEqual(rem, EXPECTED)

    @patch("safety.safety.compute_sec_ver")
    @patch("safety.safety.precompute_remediations")
    def test_calculate_remediations(self, precompute_remediations, compute_sec_ver):
        test_filename = os.path.join(self.dirname, "test_db/insecure_full_affected_versions.json")
        db_full = None
        with open(test_filename) as f:
            db_full = json.loads(f.read())
        self.assertIsNotNone(db_full)

        numpy_pkg = {'name': 'numpy', 'version': '1.22.0', 'secure_versions': ['1.22.3'],
                     'insecure_versions': ['1.22.2', '1.22.1', '1.22.0', '1.22.0rc3', '1.21.5']}

        vulns = [
            get_vulnerability(pkg_kwargs=numpy_pkg, vuln_kwargs={'affected_versions': ['1.22.0', '1.21.5']}),
            get_vulnerability(pkg_kwargs=numpy_pkg,
                              vuln_kwargs={'affected_versions': ['1.22.0', '1.22.0rc3']}),
            get_vulnerability(vuln_kwargs={'ignored': True}, pkg_kwargs={'name': 'django'})
        ]

        remediations = calculate_remediations(vulns, db_full)

        precompute_remediations.assert_called()
        compute_sec_ver.assert_called()

    def test_read_vulnerabilities_decode_error(self):
        with open(os.path.join(self.dirname, "test_db", "report_invalid_decode_error.json")) as f:
            self.assertRaises(MalformedDatabase, lambda: read_vulnerabilities(f))

    @patch("json.load")
    def test_read_vulnerabilities_type_error(self, json_load):
        json_load.side_effect = TypeError('foobar')
        with open(os.path.join(self.dirname, "test_db", "report.json")) as f:
            self.assertRaises(MalformedDatabase, lambda: read_vulnerabilities(f))

    def test_read_vulnerabilities(self):
        with open(os.path.join(self.dirname, "test_db", "report.json")) as f:
            self.assertDictEqual(self.report, read_vulnerabilities(f))

    def test_review_without_recommended_fix(self):
        vulns, remediations, packages = review(report=self.report)
        self.assertListEqual(packages, list(self.report_packages.values()))
        self.assertDictEqual(remediations, self.report_remediations)
        self.assertListEqual(vulns, self.report_vulns)

    def test_report_with_recommended_fix(self):
        REMEDIATIONS_WITH_FIX = {'django': {
            '==4.0.1': {'version': '4.0.1', 'vulnerabilities_found': 4,
                        'requirement': SafetyRequirement('django==4.0.1'),
                        'other_recommended_versions': ['2.2.28', '3.2.13'],
                        'recommended_version': parse('4.0.4'),
                        'more_info_url': 'https://pyup.io/packages/pypi/django/?from=4.0.1&to=4.0.4'}}}

        with open(os.path.join(self.dirname, "test_db", "report_with_recommended_fix.json")) as f:
            vulns, remediations, packages = review(report=read_vulnerabilities(f))
            self.assertDictEqual(remediations, REMEDIATIONS_WITH_FIX)



import abc
import os.path
import re
from pathlib import Path
from sys import platform


from safety.tool.pip import Pip
from safety.tool.poetry import Poetry


from typing import Any, TYPE_CHECKING, Optional

from safety.tool.uv.main import Uv

if TYPE_CHECKING:
    pass


def is_os_supported():
    return platform in ["linux", "linux2", "darwin", "win32"]


class BuildFileConfigurator(abc.ABC):
    @abc.abstractmethod
    def is_supported(self, file: Path) -> bool:
        """
        Returns whether a specific file is supported by this class.
        Args:
            file (str): The file to check.
        Returns:
            bool: Whether the file is supported by this class.
        """
        pass

    @abc.abstractmethod
    def configure(
        self, file: Path, org_slug: Optional[str], project_id: Optional[str]
    ) -> Optional[Path]:
        """
        Configures specific file.
        Args:
            file (str): The file to configure.
            org_slug (str): The organization slug.
            project_id (str): The project identifier.
        """
        pass


class PipRequirementsConfigurator(BuildFileConfigurator):
    __file_name_pattern = re.compile("^([a-zA-Z_-]+)?requirements([a-zA-Z_-]+)?.txt$")

    def is_supported(self, file: Path) -> bool:
        return self.__file_name_pattern.match(os.path.basename(file)) is not None

    def configure(
        self, file: Path, org_slug: Optional[str], project_id: Optional[str]
    ) -> None:
        Pip.configure_requirements(file, org_slug, project_id)


class PoetryPyprojectConfigurator(BuildFileConfigurator):
    __file_name_pattern = re.compile("^pyproject.toml$")

    def is_supported(self, file: Path) -> bool:
        return self.__file_name_pattern.match(
            os.path.basename(file)
        ) is not None and Poetry.is_poetry_project_file(file)

    def configure(
        self, file: Path, org_slug: Optional[str], project_id: Optional[str]
    ) -> Optional[Path]:
        if self.is_supported(file):
            return Poetry.configure_pyproject(file, org_slug, project_id)  # type: ignore
        return None


# TODO: Review if we should move this/hook up this into interceptors.
class ToolConfigurator(abc.ABC):
    @abc.abstractmethod
    def configure(self, org_slug: Optional[str]) -> Any:
        """
        Configures specific tool.
        Args:
            org_slug (str): The organization slug.
        """
        pass

    @abc.abstractmethod
    def reset(self) -> None:
        """
        Resets specific tool.
        """
        pass


class PipConfigurator(ToolConfigurator):
    def configure(self, org_slug: Optional[str]) -> Optional[Path]:
        return Pip.configure_system(org_slug)

    def reset(self) -> None:
        Pip.reset_system()


class PoetryConfigurator(ToolConfigurator):
    """
    Configures poetry system is not supported.
    """

    def configure(self, org_slug: Optional[str]) -> Optional[Path]:
        return None

    def reset(self) -> None:
        return None


class UvConfigurator(ToolConfigurator):
    def configure(self, org_slug: Optional[str]) -> Optional[Path]:
        return Uv.configure_system(org_slug)

    def reset(self) -> None:
        Uv.reset_system()


class UvPyprojectConfigurator(BuildFileConfigurator):
    __file_name_pattern = re.compile("^uv.lock$")

    def is_supported(self, file: Path) -> bool:
        return (
            self.__file_name_pattern.match(os.path.basename(file)) is not None
            and Path("pyproject.toml").exists()
        )

    def configure(
        self, file: Path, org_slug: Optional[str], project_id: Optional[str]
    ) -> Optional[Path]:
        if self.is_supported(file):
            return Uv.configure_pyproject(Path("pyproject.toml"), org_slug, project_id)
        return None



from typing import TYPE_CHECKING, Any, Optional
from typing_extensions import Annotated
from pydantic import ConfigDict
from safety_schemas.models.events import EventTypeBase, PayloadBase


if TYPE_CHECKING:
    pass


class InternalEventType(EventTypeBase):
    """
    Internal event types.
    """

    CLOSE_RESOURCES = "com.safetycli.close_resources"
    FLUSH_SECURITY_TRACES = "com.safetycli.flush_security_traces"
    EVENT_BUS_READY = "com.safetycli.event_bus_ready"


class InternalPayload(PayloadBase):
    ctx: Optional[Annotated[Any, "CustomContext"]] = None

    model_config = ConfigDict(extra="allow")



import json
from collections import namedtuple
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, List, Optional, Set, Union, Dict

from dparse.dependencies import Dependency
from dparse import parse, filetypes
from packaging.specifiers import SpecifierSet
from packaging.requirements import Requirement
from packaging.utils import canonicalize_name
from packaging.version import parse as parse_version

from packaging.version import Version

from safety.errors import InvalidRequirementError

try:
    from packaging.version import LegacyVersion as legacyType
except ImportError:
    legacyType = None

from .requirements import is_pinned_requirement

class DictConverter(object):
    """
    A class to convert objects to dictionaries.
    """

    def to_dict(self, **kwargs: Any) -> Dict:
        pass


announcement_nmt = namedtuple('Announcement', ['type', 'message'])
remediation_nmt = namedtuple('Remediation', ['Package', 'closest_secure_version', 'secure_versions',
                                             'latest_package_version'])
cve_nmt = namedtuple('Cve', ['name', 'cvssv2', 'cvssv3'])
severity_nmt = namedtuple('Severity', ['source', 'cvssv2', 'cvssv3'])
vulnerability_nmt = namedtuple('Vulnerability',
                               ['vulnerability_id', 'package_name', 'pkg', 'ignored', 'ignored_reason', 'ignored_expires',
                                'vulnerable_spec', 'all_vulnerable_specs', 'analyzed_version', 'analyzed_requirement',
                                'advisory', 'is_transitive', 'published_date', 'fixed_versions',
                                'closest_versions_without_known_vulnerabilities', 'resources', 'CVE', 'severity',
                                'affected_versions', 'more_info_url'])
RequirementFile = namedtuple('RequirementFile', ['path'])


class SafetyRequirement(Requirement):
    """
    A subclass of Requirement that includes additional attributes and methods for safety requirements.
    """
    def __init__(self, requirement: Union[str, Dependency], found: Optional[str] = None) -> None:
        """
        Initialize a SafetyRequirement.

        Args:
            requirement (Union[str, Dependency]): The requirement as a string or Dependency object.
            found (Optional[str], optional): Where the requirement was found. Defaults to None.

        Raises:
            InvalidRequirementError: If the requirement cannot be parsed.
        """
        dep = requirement

        if isinstance(requirement, str):
            deps = parse(requirement, file_type=filetypes.requirements_txt).dependencies
            dep = deps[0] if deps else None

        if not dep:
            raise InvalidRequirementError(line=str(requirement))

        raw_line = dep.line
        to_parse = dep.line
        # Hash and comments are only a pip feature, so removing them.
        if '#' in to_parse:
            to_parse = dep.line.split('#')[0]

        for req_hash in dep.hashes:
            to_parse = to_parse.replace(req_hash, '')

        to_parse = to_parse.replace('\\', '').rstrip()

        try:
            # Try to build a PEP Requirement from the cleaned line
            super(SafetyRequirement, self).__init__(to_parse)
        except Exception:
            raise InvalidRequirementError(line=requirement.line if isinstance(requirement, Dependency) else requirement)

        self.raw = raw_line
        self.found = found

    def __eq__(self, other: Any) -> bool:
        return str(self) == str(other)

    def to_dict(self, **kwargs: Any) -> Dict:
        """
        Convert the requirement to a dictionary.

        Args:
            **kwargs: Additional keyword arguments.

        Returns:
            dict: The dictionary representation of the requirement.
        """
        specifier_obj = self.specifier
        if not "specifier_obj" in kwargs:
            specifier_obj = str(self.specifier)

        return   {
                        'raw': self.raw,
                        'extras': list(self.extras),
                        'marker': str(self.marker) if self.marker else None,
                        'name': self.name,
                        'specifier': specifier_obj,
                        'url': self.url,
                        'found': self.found
                    }


@dataclass
class Package(DictConverter):
    """
    A class representing a package.
    """
    name: str
    version: Optional[str]
    requirements: List[SafetyRequirement]
    found: Optional[str] = None
    absolute_path: Optional[str] = None
    insecure_versions: List[str] = field(default_factory=lambda: [])
    secure_versions: List[str] = field(default_factory=lambda: [])
    latest_version_without_known_vulnerabilities: Optional[str] = None
    latest_version: Optional[str] = None
    more_info_url: Optional[str] = None

    def has_unpinned_req(self) -> bool:
        """
        Check if the package has unpinned requirements.

        Returns:
            bool: True if there are unpinned requirements, False otherwise.
        """
        for req in self.requirements:
            if not is_pinned_requirement(req.specifier):
                return True
        return False

    def get_unpinned_req(self) -> filter:
        """
        Get the unpinned requirements.

        Returns:
            filter: A filter object with the unpinned requirements.
        """
        return filter(lambda r: not is_pinned_requirement(r.specifier), self.requirements)

    def filter_by_supported_versions(self, versions: List[str]) -> List[str]:
        """
        Filter the versions by supported versions.

        Args:
            versions (List[str]): The list of versions.

        Returns:
            List[str]: The list of supported versions.
        """
        allowed = []

        for version in versions:
            try:
                parse_version(version)
                allowed.append(version)
            except Exception:
                pass

        return allowed

    def get_versions(self, db_full: Dict) -> Set[str]:
        """
        Get the versions from the database.

        Args:
            db_full (Dict): The full database.

        Returns:
            Set[str]: The set of versions.
        """
        pkg_meta = db_full.get('meta', {}).get('packages', {}).get(self.name, {})
        versions = self.filter_by_supported_versions(
            pkg_meta.get("insecure_versions", []) + pkg_meta.get("secure_versions", []))
        return set(versions)

    def refresh_from(self, db_full: Dict) -> None:
        """
        Refresh the package information from the database.

        Args:
            db_full (Dict): The full database.
        """
        base_domain = db_full.get('meta', {}).get('base_domain')
        pkg_meta = db_full.get('meta', {}).get('packages', {}).get(canonicalize_name(self.name), {})

        kwargs = {'insecure_versions': self.filter_by_supported_versions(pkg_meta.get("insecure_versions", [])),
                  'secure_versions': self.filter_by_supported_versions(pkg_meta.get("secure_versions", [])),
                  'latest_version_without_known_vulnerabilities': pkg_meta.get("latest_secure_version",
                                                                               None),
                  'latest_version': pkg_meta.get("latest_version", None),
                  'more_info_url': f"{base_domain}{pkg_meta.get('more_info_path', '')}"}

        self.update(kwargs)

    def to_dict(self, **kwargs: Any) -> Dict:
        """
        Convert the package to a dictionary.

        Args:
            **kwargs: Additional keyword arguments.

        Returns:
            dict: The dictionary representation of the package.
        """
        if kwargs.get('short_version', False):
            return {
                'name': self.name,
                'version': self.version,
                'requirements': self.requirements
            }

        return {'name': self.name,
                'version': self.version,
                'requirements': self.requirements,
                'found': None,
                'insecure_versions': self.insecure_versions,
                'secure_versions': self.secure_versions,
                'latest_version_without_known_vulnerabilities': self.latest_version_without_known_vulnerabilities,
                'latest_version': self.latest_version,
                'more_info_url': self.more_info_url
                }

    def update(self, new: Dict) -> None:
        """
        Update the package attributes with new values.

        Args:
            new (Dict): The new attribute values.
        """
        for key, value in new.items():
            if hasattr(self, key):
                setattr(self, key, value)


class Announcement(announcement_nmt):
    """
    A class representing an announcement.
    """
    pass


class Remediation(remediation_nmt, DictConverter):
    """
    A class representing a remediation.
    """

    def to_dict(self) -> Dict:
        """
        Convert the remediation to a dictionary.

        Returns:
            Dict: The dictionary representation of the remediation.
        """
        return {'package': self.Package.name,
                'closest_secure_version': self.closest_secure_version,
                'secure_versions': self.secure_versions,
                'latest_package_version': self.latest_package_version
                }


@dataclass
class Fix:
    """
    A class representing a fix.
    """
    dependency: Any = None
    previous_version: Any = None
    previous_spec: Optional[str] = None
    other_options: List[str] = field(default_factory=lambda: [])
    updated_version: Any = None
    update_type: str = ""
    package: str = ""
    status: str = ""
    applied_at: str = ""
    fix_type: str = ""
    more_info_url: str = ""


class CVE(cve_nmt, DictConverter):
    """
    A class representing a CVE.
    """

    def to_dict(self) -> Dict:
        """
        Convert the CVE to a dictionary.

        Returns:
            Dict: The dictionary representation of the CVE.
        """
        return {'name': self.name, 'cvssv2': self.cvssv2, 'cvssv3': self.cvssv3}


class Severity(severity_nmt, DictConverter):
    """
    A class representing the severity of a vulnerability.
    """
    def to_dict(self) -> Dict:
        """
        Convert the severity to a dictionary.

        Returns:
            Dict: The dictionary representation of the severity.
        """
        result = {'severity': {'source': self.source}}

        result['severity']['cvssv2'] = self.cvssv2
        result['severity']['cvssv3'] = self.cvssv3

        return result


class SafetyEncoder(json.JSONEncoder):
    """
    A custom JSON encoder for Safety related objects.
    """
    def default(self, value: Any) -> Any:
        """
        Override the default method to handle custom objects.

        Args:
            value (Any): The value to encode.

        Returns:
            Any: The encoded value.
        """
        if isinstance(value, SafetyRequirement):
            return value.to_dict()
        elif isinstance(value, Version) or (legacyType and isinstance(value, legacyType)):
            return str(value)
        else:
            return super().default(value)


class Vulnerability(vulnerability_nmt):
    """
    A class representing a vulnerability.
    """

    def to_dict(self) -> Dict:
        """
        Convert the vulnerability to a dictionary.

        Returns:
            Dict: The dictionary representation of the vulnerability.
        """
        empty_list_if_none = ['fixed_versions', 'closest_versions_without_known_vulnerabilities', 'resources']
        result = {
        }

        ignore = ['pkg']

        for field, value in zip(self._fields, self):
            if field in ignore:
                continue

            if value is None and field in empty_list_if_none:
                value = []

            if isinstance(value, set):
                result[field] = list(value)
            elif isinstance(value, CVE):
                val = None
                if value.name.startswith("CVE"):
                    val = value.name
                result[field] = val
            elif isinstance(value, DictConverter):
                result.update(value.to_dict())
            elif isinstance(value, SpecifierSet) or isinstance(value, datetime):
                result[field] = str(value)
            else:
                result[field] = value

        return result

    def get_advisory(self) -> str:
        """
        Get the advisory for the vulnerability.

        Returns:
            str: The advisory text.
        """
        return self.advisory.replace('\r', '') if self.advisory else "No advisory found for this vulnerability."

    def to_model_dict(self) -> Dict:
        """
        Convert the vulnerability to a dictionary for the model.

        Returns:
            Dict: The dictionary representation of the vulnerability for the model.
        """
        try:
            affected_spec = next(iter(self.vulnerable_spec))
        except Exception:
            affected_spec = ""

        repr = {
            "id": self.vulnerability_id,
            "package_name": self.package_name,
            "vulnerable_spec": affected_spec,
            "analyzed_specification": self.analyzed_requirement.raw
        }

        if self.ignored:
            repr["ignored"] = {"reason": self.ignored_reason,
                               "expires": self.ignored_expires}

        return repr


@dataclass
class Safety:
    """
    A class representing Safety settings.
    """
    client: Any
    keys: Any




from .emission import (
    emit_command_error,
    emit_command_executed,
    emit_firewall_disabled,
    emit_diff_operations,
    emit_firewall_configured,
    emit_tool_command_executed,
    emit_firewall_heartbeat,
    emit_init_started,
    emit_auth_started,
    emit_auth_completed,
)

from .creation import (
    create_internal_event,
    InternalEventType,
    InternalPayload,
)

__all__ = [
    "emit_command_error",
    "emit_command_executed",
    "emit_firewall_disabled",
    "create_internal_event",
    "InternalEventType",
    "InternalPayload",
    "emit_firewall_configured",
    "emit_diff_operations",
    "emit_init_started",
    "emit_auth_started",
    "emit_auth_completed",
    "emit_tool_command_executed",
    "emit_firewall_heartbeat",
]



import logging
from pathlib import Path
import re
import shutil
import tempfile
from typing import List
from .base import CommandInterceptor, Tool
from .types import InterceptorType

from safety.constants import USER_CONFIG_DIR


logger = logging.getLogger(__name__)


class UnixAliasInterceptor(CommandInterceptor):
    def __init__(self):
        super().__init__(InterceptorType.UNIX_ALIAS)
        self.user_rc_paths: List[Path] = self._get_user_rc_paths()
        self.custom_rc_path = self._get_custom_rc_path()
        self.legacy_user_rc_paths = [Path.home() / ".profile"]

        # Update these markers could be a breaking change; be careful to handle
        # backward compatibility
        self.marker_start = "# >>> Safety >>>"
        self.marker_end = "# <<< Safety <<<"

    def _get_user_rc_paths(self) -> List[Path]:
        """
        We support the following shells:
        * Zsh
        * Bash
        """
        zsh_paths = [Path.home() / ".zshrc"]
        # .bash_profile is added for max compatibility on macOS
        bash_paths = [Path.home() / ".bashrc", Path.home() / ".bash_profile"]

        return zsh_paths + bash_paths

    def _get_custom_rc_path(self) -> Path:
        return USER_CONFIG_DIR / ".safety_profile"

    def _backup_file(self, path: Path) -> None:
        """
        Create backup of file if it exists
        """
        if path.exists():
            backup_path = path.with_suffix(".backup")
            shutil.copy2(path, backup_path)

    def _generate_user_rc_content(self) -> str:
        """
        Generate the content to be added to user's rc.

        Example:
        ```
        # >>> Safety >>>
        [ -f "$HOME/.safety/.safety_profile" ] && . "$HOME/.safety/.safety_profile"
        # <<< Safety <<<
        ```
        """
        lines = (
            self.marker_start,
            f'[ -f "{self.custom_rc_path}" ] && . "{self.custom_rc_path}"',
            self.marker_end,
        )
        return "\n".join(lines) + "\n"

    def _is_configured(self, user_rc_path: Path) -> bool:
        """
        Check if the configuration block exists in user's rc file
        """
        try:
            if not user_rc_path.exists():
                return False

            content = user_rc_path.read_text()
            return self.marker_start in content and self.marker_end in content

        except OSError:
            logger.info("Failed to read user's rc file")
            return False

    def _generate_custom_rc_content(self, aliases: List[str]) -> str:
        """
        Generate the content for the custom profile with metadata
        """
        metadata_lines = self._generate_metadata_content(prepend="#")
        aliases_lines = tuple(aliases)

        lines = (
            (self.marker_start,) + metadata_lines + aliases_lines + (self.marker_end,)
        )

        return "\n".join(lines) + "\n"

    def _ensure_source_line_in_user_rc(self) -> None:
        """
        Ensure source line exists in user's rc files

        If the source line is not present in the user's rc files, append it.
        If the user's rc files do not exist, create them.
        """
        source_line = self._generate_user_rc_content()

        for user_rc_path in self.user_rc_paths:
            if not user_rc_path.exists():
                user_rc_path.write_text(source_line)
                continue

            if not self._is_configured(user_rc_path):
                with open(user_rc_path, "a") as f:
                    f.write(source_line)

    def _batch_install_tools(self, tools: List[Tool]) -> bool:
        """
        Install aliases for multiple tools
        """
        try:
            # Generate aliases
            aliases = []
            for tool in tools:
                for binary in tool.binary_names:
                    alias_def = f'alias {binary}="safety {binary}"'
                    aliases.append(alias_def)

            if not aliases:
                return False

            # Create safety profile directory if it doesn't exist
            self.custom_rc_path.parent.mkdir(parents=True, exist_ok=True)

            # Generate new profile content
            content = self._generate_custom_rc_content(aliases)

            # Backup target files
            for f_path in self.user_rc_paths + [self.custom_rc_path]:
                self._backup_file(path=f_path)

            # Override our custom profile
            # TODO: handle exceptions
            self.custom_rc_path.write_text(content)

            # Ensure source line in user's rc files
            self._ensure_source_line_in_user_rc()

            return True

        except Exception:
            logger.exception("Failed to batch install aliases")
            return False

    def _batch_remove_tools(self, tools: List[Tool]) -> bool:
        """
        This will remove all the tools.

        NOTE: for now this does not support to remove individual tools.
        """
        try:
            # Backup target files
            for f_path in self.user_rc_paths + [self.custom_rc_path]:
                self._backup_file(path=f_path)

            for user_rc_path in self.user_rc_paths + self.legacy_user_rc_paths:
                if self._is_configured(user_rc_path):
                    temp_dir = tempfile.gettempdir()
                    temp_file = Path(temp_dir) / f"{user_rc_path.name}.tmp"

                    pattern = rf"{self.marker_start}\n.*?\{self.marker_end}\n?"

                    with open(user_rc_path, "r") as src, open(temp_file, "w") as dst:
                        content = src.read()
                        cleaned_content = re.sub(pattern, "", content, flags=re.DOTALL)
                        dst.write(cleaned_content)

                    if not temp_file.exists():
                        logger.info("Temp file is empty or invalid")
                        return False

                    shutil.move(str(temp_file), str(user_rc_path))

            self.custom_rc_path.unlink(missing_ok=True)

            return True
        except Exception as e:
            logger.exception(f"Failed to batch remove aliases: {e}")
            return False

    def _install_tool(self, tool: Tool) -> bool:
        return self._batch_install_tools([tool])

    def _remove_tool(self, tool: Tool) -> bool:
        return self._batch_remove_tools([tool])



from enum import Enum
import logging
import os
from pathlib import Path
import subprocess
from typing import TYPE_CHECKING, Optional, Tuple


from safety.scan.finder.handlers import (
    FileHandler,
    PythonFileHandler,
    SafetyProjectFileHandler,
)
from safety_schemas.models import Stage

if TYPE_CHECKING:
    from safety_schemas.models import GITModel


LOG = logging.getLogger(__name__)


class Language(str, Enum):
    """
    Enum representing supported programming languages.
    """

    python = "python"
    javascript = "javascript"
    safety_project = "safety_project"

    def handler(self) -> FileHandler:
        """
        Get the appropriate file handler for the language.

        Returns:
            FileHandler: The file handler for the language.
        """
        if self is Language.python:
            return PythonFileHandler()
        if self is Language.safety_project:
            return SafetyProjectFileHandler()

        return PythonFileHandler()


class Output(Enum):
    """
    Enum representing output formats.
    """

    json = "json"


class AuthenticationType(str, Enum):
    """
    Enum representing authentication types.
    """

    token = "token"
    api_key = "api_key"
    none = "unauthenticated"

    def is_allowed_in(self, stage: Stage = Stage.development) -> bool:
        """
        Check if the authentication type is allowed in the given stage.

        Args:
            stage (Stage): The current stage.

        Returns:
            bool: True if the authentication type is allowed, otherwise False.
        """
        if self is AuthenticationType.none:
            return False

        if stage == Stage.development and self is AuthenticationType.api_key:
            return False

        if (not stage == Stage.development) and self is AuthenticationType.token:
            return False

        return True


class GIT:
    """
    Class representing Git operations.
    """

    ORIGIN_CMD: Tuple[str, ...] = ("remote", "get-url", "origin")
    BRANCH_CMD: Tuple[str, ...] = ("symbolic-ref", "--short", "-q", "HEAD")
    TAG_CMD: Tuple[str, ...] = ("describe", "--tags", "--exact-match")
    DESCRIBE_CMD: Tuple[str, ...] = (
        "describe",
        '--match=""',
        "--always",
        "--abbrev=40",
        "--dirty",
    )
    GIT_CHECK_CMD: Tuple[str, ...] = ("rev-parse", "--is-inside-work-tree")

    def __init__(self, root: Path = Path(".")) -> None:
        """
        Initialize the GIT class with the given root directory.

        Args:
            root (Path): The root directory for Git operations.
        """
        self.git = ("git", "-C", root.resolve())

    def __run__(
        self, cmd: Tuple[str, ...], env_var: Optional[str] = None
    ) -> Optional[str]:
        """
        Run a Git command.

        Args:
            cmd (Tuple[str, ...]): The Git command to run.
            env_var (Optional[str]): An optional environment variable to check for the command result.

        Returns:
            Optional[str]: The result of the Git command, or None if an error occurred.
        """
        if env_var and os.environ.get(env_var):
            return os.environ.get(env_var)

        try:
            return (
                subprocess.run(
                    self.git + cmd, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL
                )
                .stdout.decode("utf-8")
                .strip()
            )
        except Exception as e:
            LOG.exception(e)

        return None

    def origin(self) -> Optional[str]:
        """
        Get the Git origin URL.

        Returns:
            Optional[str]: The Git origin URL, or None if an error occurred.
        """
        return self.__run__(self.ORIGIN_CMD, env_var="SAFETY_GIT_ORIGIN")

    def branch(self) -> Optional[str]:
        """
        Get the current Git branch.

        Returns:
            Optional[str]: The current Git branch, or None if an error occurred.
        """
        return self.__run__(self.BRANCH_CMD, env_var="SAFETY_GIT_BRANCH")

    def tag(self) -> Optional[str]:
        """
        Get the current Git tag.

        Returns:
            Optional[str]: The current Git tag, or None if an error occurred.
        """
        return self.__run__(self.TAG_CMD, env_var="SAFETY_GIT_TAG")

    def describe(self) -> Optional[str]:
        """
        Get the Git describe output.

        Returns:
            Optional[str]: The Git describe output, or None if an error occurred.
        """
        return self.__run__(self.DESCRIBE_CMD)

    def dirty(self, raw_describe: str) -> bool:
        """
        Check if the working directory is dirty.

        Args:
            raw_describe (str): The raw describe output.

        Returns:
            bool: True if the working directory is dirty, otherwise False.
        """
        if (is_dirty := os.environ.get("SAFETY_GIT_DIRTY")) and is_dirty in ["0", "1"]:
            return bool(int(is_dirty))

        return raw_describe.endswith("-dirty")

    def commit(self, raw_describe: str) -> Optional[str]:
        """
        Get the current Git commit hash.

        Args:
            raw_describe (str): The raw describe output.

        Returns:
            Optional[str]: The current Git commit hash, or None if an error occurred.
        """
        if os.environ.get("SAFETY_GIT_COMMIT"):
            return os.environ.get("SAFETY_GIT_COMMIT")

        try:
            return raw_describe.split("-dirty")[0]
        except Exception:
            pass

    def is_git(self) -> bool:
        """
        Check if the current directory is a Git repository.

        Returns:
            bool: True if the current directory is a Git repository, otherwise False.
        """
        result = self.__run__(self.GIT_CHECK_CMD)

        if result == "true":
            return True

        return False

    def build_git_data(self) -> Optional["GITModel"]:
        """
        Build a GITModel object with Git data.

        Returns:
            Optional[GITModel]: The GITModel object with Git data, or None if the directory is not a Git repository.
        """
        from safety_schemas.models import GITModel

        if self.is_git():
            raw_describe = self.describe()
            commit = None
            dirty = False

            # TODO: describe fails when there are not commits,
            # GitModel needs to support this case too
            if raw_describe:
                commit = self.commit(raw_describe)
                dirty = self.dirty(raw_describe)
            return GITModel(
                branch=self.branch(),
                tag=self.tag(),
                commit=commit,
                dirty=dirty,
                origin=self.origin(),
            )

        return None



import unittest
from unittest.mock import patch, MagicMock, mock_open

from safety.alerts import utils
from tests.test_cli import get_vulnerability


class TestUtils(unittest.TestCase):
    def setUp(self):
        self.pkg = "requests"
        severity_kwargs = [
            {'name': 'CVE-2014-0474', 'cvssv2': None, 'cvssv3': None},
            {'name': 'CVE-2019-3498',
             'cvssv2': {"base_score": 4.3, "impact_score": 2.9,
                        "vector_string": "AV:N/AC:M/Au:N/C:N/I:P/A:N"},
             'cvssv3': {"base_score": 6.5, "impact_score": 3.6, "base_severity": "MEDIUM",
                        "vector_string": "CVSS:3.0/AV:N/AC:L/PR:N/UI:R/S:U/C:N/I:H/A:N"}},
            {'name': 'CVE-2021-29601', 'cvssv2': None,
             'cvssv3': {"base_score": 7.1, "impact_score": 5.2, "base_severity": "HIGH",
                        "vector_string": "CVSS:3.1/AV:L/AC:L/PR:L/UI:N/S:U/C:N/I:H/A:H"}},
            {'name': 'CVE-2021-29601', 'cvssv2': {"base_score": 4.3, "impact_score": 2.9,
                                                  "vector_string": "AV:N/AC:M/Au:N/C:N/I:P/A:N"},
             'cvssv3': None}
        ]
        self.vulns = list(map(lambda cve_kwargs: get_vulnerability(cve_kwargs=cve_kwargs).to_dict(), severity_kwargs))
        self.unpinned_remediation = {
            "version": None,
            "recommended_version": "2.22.0",
            "requirement": {"name": "requests", "specifier": ">=2.21.0,<2.22.0"}
        }
        self.pinned_remediation = {
            "version": '2.21.0',
            "recommended_version": "2.22.0",
            "requirement": {"name": "requests", "specifier": "==2.21.0"}
        }
        self.api_key = "api_key"

    def test_highest_base_score_with_an_unknown_cvssv3(self):
        score = utils.highest_base_score(self.vulns)
        # TODO: This needs validation with users, it at least a CVSSv3 is None, then it returns 10
        self.assertEqual(score, 10)

    def test_highest_base_score(self):
        cvssv3_vulns = filter(lambda v: (v.get('severity', {}) or {}).get('cvssv3', {}), self.vulns)
        score = utils.highest_base_score(cvssv3_vulns)
        self.assertEqual(score, 7.1)

    def test_generate_branch_name_unpinned_remediation(self):
        branch_name = utils.generate_branch_name(self.pkg, self.unpinned_remediation)
        self.assertEqual(branch_name, "requests/>=2.21.0,<2.22.0/2.22.0")

    def test_generate_branch_name_pinned_remediation(self):
        branch_name = utils.generate_branch_name(self.pkg, self.pinned_remediation)
        self.assertEqual(branch_name, "requests/==2.21.0/2.22.0")

    def test_generate_issue_title_unpinned_remediation(self):
        issue_title = utils.generate_issue_title(self.pkg, self.unpinned_remediation)
        self.assertEqual(issue_title, "Security Vulnerability in requests>=2.21.0,<2.22.0")

    def test_generate_issue_title_pinned_remediation(self):
        issue_title = utils.generate_issue_title(self.pkg, self.pinned_remediation)
        self.assertEqual(issue_title, "Security Vulnerability in requests==2.21.0")

    def test_generate_title_pinned_remediation(self):
        title = utils.generate_title(self.pkg, self.pinned_remediation, [self.vulns[0]])
        self.assertEqual(title, "Update requests from 2.21.0 to 2.22.0 to fix 1 vulnerability")

        title = utils.generate_title(self.pkg, self.pinned_remediation, self.vulns)
        self.assertEqual(title, "Update requests from 2.21.0 to 2.22.0 to fix 4 vulnerabilities")

    def test_generate_title_unpinned_remediation(self):
        title = utils.generate_title(self.pkg, self.unpinned_remediation, [self.vulns[0]])
        self.assertEqual(title, "Update requests from >=2.21.0,<2.22.0 to 2.22.0 to fix 1 vulnerability")

        title = utils.generate_title(self.pkg, self.unpinned_remediation, self.vulns)
        self.assertEqual(title, "Update requests from >=2.21.0,<2.22.0 to 2.22.0 to fix 4 vulnerabilities")

    def test_cvss3_score_to_label_low_score(self):
        score = 1.5
        expected_label = 'low'
        self.assertEqual(utils.cvss3_score_to_label(score), expected_label)

    def test_cvss3_score_to_label_medium_score(self):
        score = 5.2
        expected_label = 'medium'
        self.assertEqual(utils.cvss3_score_to_label(score), expected_label)

    def test_cvss3_score_to_label_high_score(self):
        score = 7.8
        expected_label = 'high'
        self.assertEqual(utils.cvss3_score_to_label(score), expected_label)

    def test_cvss3_score_to_label_critical_score(self):
        score = 10.0
        expected_label = 'critical'
        self.assertEqual(utils.cvss3_score_to_label(score), expected_label)

    def test_cvss3_score_to_label_invalid_score(self):
        score = -1.0
        expected_label = None
        self.assertEqual(utils.cvss3_score_to_label(score), expected_label)

        score = 11.0
        expected_label = None
        self.assertEqual(utils.cvss3_score_to_label(score), expected_label)



# type: ignore
import unittest
from unittest.mock import MagicMock

from safety_schemas.models.events import EventType
from safety.events.types import InternalEventType
from safety.firewall.events.utils import register_event_handlers
from safety.firewall.events.handlers import HeartbeatInspectionEventHandler
from safety.models.obj import SafetyCLI


class TestFirewallEventsUtils(unittest.TestCase):
    """
    Test cases for firewall events utils functions.
    """

    def setUp(self):
        self.event_bus = MagicMock()
        self.safety_cli = MagicMock(spec=SafetyCLI)
        self.safety_cli.security_events_handler = MagicMock()

    def test_register_event_handlers_with_security_events_handler(self):
        """
        Test register_event_handlers when security_events_handler is available.
        """
        register_event_handlers(self.event_bus, self.safety_cli)

        self.event_bus.subscribe.assert_any_call(
            [InternalEventType.EVENT_BUS_READY],
            self.event_bus.subscribe.call_args_list[0][0][
                1
            ],  # Get the handler from the first call
        )

        self.event_bus.subscribe.assert_any_call(
            [
                EventType.FIREWALL_CONFIGURED,
                EventType.FIREWALL_HEARTBEAT,
                EventType.FIREWALL_DISABLED,
                EventType.PACKAGE_INSTALLED,
                EventType.PACKAGE_UNINSTALLED,
                EventType.PACKAGE_UPDATED,
                EventType.TOOL_COMMAND_EXECUTED,
                EventType.INIT_STARTED,
                EventType.FIREWALL_SETUP_RESPONSE_CREATED,
                EventType.FIREWALL_SETUP_COMPLETED,
                EventType.CODEBASE_DETECTION_STATUS,
                EventType.CODEBASE_SETUP_RESPONSE_CREATED,
                EventType.CODEBASE_SETUP_COMPLETED,
                EventType.INIT_SCAN_COMPLETED,
                EventType.INIT_EXITED,
            ],
            self.safety_cli.security_events_handler,
        )

        self.assertEqual(self.event_bus.subscribe.call_count, 2)

    def test_register_event_handlers_without_security_events_handler(self):
        """
        Test register_event_handlers when security_events_handler is None.
        """
        self.safety_cli.security_events_handler = None

        register_event_handlers(self.event_bus, self.safety_cli)

        self.assertEqual(self.event_bus.subscribe.call_count, 1)

        call_args = self.event_bus.subscribe.call_args
        events_arg = call_args[0][0]
        handler_arg = call_args[0][1]

        self.assertEqual(events_arg, [InternalEventType.EVENT_BUS_READY])
        self.assertIsInstance(handler_arg, HeartbeatInspectionEventHandler)
        self.assertEqual(handler_arg.event_bus, self.event_bus)




import os
from pathlib import Path
from typing import Optional, Tuple
import typer
from safety.scan.models import ScanExport, ScanOutput

from safety_schemas.models import AuthenticationType


MISSING_SPDX_EXTENSION_MSG = "spdx extra is not installed, please install it with: pip install safety[spdx]"


def raise_if_not_spdx_extension_installed() -> None:
    """
    Raises an error if the spdx extension is not installed.
    """
    try:
        import spdx_tools.spdx
    except Exception as e:
        raise typer.BadParameter(MISSING_SPDX_EXTENSION_MSG)


def save_as_callback(save_as: Optional[Tuple[ScanExport, Path]]) -> Tuple[Optional[str], Optional[Path]]:
    """
    Callback function to handle save_as parameter and validate if spdx extension is installed.

    Args:
        save_as (Optional[Tuple[ScanExport, Path]]): The export type and path.

    Returns:
        Tuple[Optional[str], Optional[Path]]: The validated export type and path.
    """
    export_type, export_path = save_as if save_as else (None, None)

    if ScanExport.is_format(export_type, ScanExport.SPDX):
        raise_if_not_spdx_extension_installed()

    return (export_type.value, export_path) if export_type and export_path else (export_type, export_path)

def output_callback(output: ScanOutput) -> str:
    """
    Callback function to handle output parameter and validate if spdx extension is installed.

    Args:
        output (ScanOutput): The output format.

    Returns:
        str: The validated output format.
    """
    if ScanOutput.is_format(output, ScanExport.SPDX):
        raise_if_not_spdx_extension_installed()

    return output.value


def fail_if_not_allowed_stage(ctx: typer.Context):
    """
    Fail the command if the authentication type is not allowed in the current stage.

    Args:
        ctx (typer.Context): The context of the Typer command.
    """
    if ctx.resilient_parsing:
        return

    stage = ctx.obj.auth.stage
    auth_type: AuthenticationType = ctx.obj.auth.client.get_authentication_type()

    if os.getenv("SAFETY_DB_DIR"):
        return

    if not auth_type.is_allowed_in(stage):
        raise typer.BadParameter(f"'{auth_type.value}' auth type isn't allowed with " \
                                 f"the '{stage}' stage.")



import logging
from pathlib import Path
import sys
import threading
import time
from typing import (
    TYPE_CHECKING,
    Optional,
    Iterator,
)

from rich.live import Live
from rich.padding import Padding
from rich.panel import Panel
from rich.prompt import Prompt
from rich.syntax import Syntax
from rich.table import Table
from rich.text import Text

from safety.events.utils.emission import (
    emit_codebase_detection_status,
    emit_codebase_setup_completed,
    emit_init_exited,
    emit_init_scan_completed,
)
from safety.init.models import StepTracker
from safety_schemas.models.events.types import ToolType


from .render import (
    ask_codebase_setup,
    ask_continue,
    ask_firewall_setup,
    load_emoji,
    progressive_print,
    render_header,
    typed_print,
)
from safety.scan.init_scan import start_scan
from ..cli_util import (
    CommandType,
    FeatureType,
    SafetyCLICommand,
    SafetyCLISubGroup,
)
from safety.error_handlers import handle_cmd_exception
import typer


from safety.init.constants import (
    MSG_ANALYZE_CODEBASE_TITLE,
    MSG_CODEBASE_FAILED_TO_SCAN,
    MSG_CODEBASE_NOT_CONFIGURED,
    MSG_CODEBASE_URL_DESCRIPTION,
    MSG_COMPLETE_SECURED,
    MSG_COMPLETE_TOOL_SECURED,
    MSG_FIREWALL_UNINSTALL,
    MSG_LAST_MANUAL_STEP,
    MSG_NO_VULNERABILITIES_FOUND,
    MSG_NO_VULNS_CODEBASE_URL_DESCRIPTION,
    MSG_OPEN_DASHBOARD_PROMPT,
    MSG_SETUP_CODEBASE_NO_PROJECT,
    MSG_SETUP_COMPLETE_SUBTITLE,
    MSG_SETUP_COMPLETE_TITLE,
    MSG_SETUP_INCOMPLETE,
    MSG_SETUP_NEXT_STEPS,
    MSG_SETUP_NEXT_STEPS_MANUAL_STEP,
    MSG_SETUP_NEXT_STEPS_NO_PROJECT,
    MSG_SETUP_NEXT_STEPS_NO_VULNS,
    MSG_SETUP_NEXT_STEPS_SUBTITLE,
    MSG_SETUP_PACKAGE_FIREWALL_DESCRIPTION,
    MSG_SETUP_PACKAGE_FIREWALL_TITLE,
    MSG_SETUP_CODEBASE_DESCRIPTION,
    MSG_SETUP_CODEBASE_TITLE,
    CODEBASE_INIT_CMD_NAME,
    CODEBASE_INIT_HELP,
    CODEBASE_INIT_DIRECTORY_HELP,
    MSG_TOOLS_NOT_CONFIGURED,
    MSG_WELCOME_TITLE,
    MSG_WELCOME_DESCRIPTION,
)
from safety.init.main import create_project, launch_auth_if_needed, setup_firewall
from safety.console import main_console as console
from ..tool.main import (
    configure_local_directory,
    find_local_tool_files,
)

from ..constants import CONTEXT_COMMAND_TYPE, CONTEXT_FEATURE_TYPE

from safety.decorators import notify
from safety.events.utils import emit_firewall_configured, emit_init_started
from safety_schemas.models.events.payloads import AliasConfig, IndexConfig, InitExitStep


if TYPE_CHECKING:
    import typer
    from safety.scan.init_scan import ScanResult

try:
    from typing import Annotated  # type: ignore
except ImportError:
    from typing_extensions import Annotated

init_app = typer.Typer(rich_markup_mode="rich", cls=SafetyCLISubGroup)

logger = logging.getLogger(__name__)


class InitScanState:
    """
    Class to track scan state for vulnerability scans

    Attributes:
        scan_id: ID of the scan
        dependencies: Number of dependencies found
        critical: Count of critical vulnerabilities
        high: Count of high severity vulnerabilities
        medium: Count of medium severity vulnerabilities
        low: Count of low severity vulnerabilities
        fixes: Number of fixes available
        fixed_vulns: Number of vulnerabilities with fixes
        url: URL to view the scan results
        completed: Whether the scan has completed
        progress: Percentage progress of the scan
        status_message: Current status message from the scanner
        status_action: Current action being performed by the scanner
        current_file: Current file being processed
    """

    def __init__(self):
        self.scan_id: Optional[str] = None
        self.dependencies: int = 0
        self.critical: int = 0
        self.high: int = 0
        self.medium: int = 0
        self.low: int = 0
        self.others: int = 0
        self.vulns_count: int = 0
        self.fixes: int = 0
        self.fixed_vulns: int = 0
        self.codebase_url: Optional[str] = None
        self.completed: bool = False
        self.progress: int = 0
        self.status_message: Optional[str] = None
        self.status_action: Optional[str] = None
        self.current_file: Optional[str] = None


def generate_summary(state: InitScanState, spinner_phase=0):
    """
    Generate the summary text based on current scan state
    """
    text = Text()

    spinner = "⠋⠙⠹⠸⠼⠴⠦⠧⠇⠏"

    text.append(
        f"Tested {state.dependencies} dependenc{'y' if state.dependencies == 1 else 'ies'} for security issues\n"
    )
    text.append("\nFound:\n\n")

    categories = [
        {
            "name": "CRITICAL",
            "icon": "🔥",
            "style": "bold red",
            "dim_style": "dim red",
            "count_attr": "critical",
            "spinner_offset": 0,
        },
        {
            "name": "HIGH",
            "icon": "🟡",
            "style": "bold yellow",
            "dim_style": "dim yellow",
            "count_attr": "high",
            "spinner_offset": 2,
        },
        {
            "name": "MEDIUM",
            "icon": "!!",
            "style": "yellow",
            "dim_style": "dim yellow",
            "count_attr": "medium",
            "spinner_offset": 4,
        },
        {
            "name": "LOW",
            "icon": "ℹ️ ",
            "style": "bold blue",
            "dim_style": "dim blue",
            "count_attr": "low",
            "spinner_offset": 6,
        },
        {
            "name": "OTHER",
            "icon": "**",
            "style": "blue",
            "dim_style": "dim blue",
            "count_attr": "others",
            "spinner_offset": 8,
        },
    ]

    # No vulnerabilities case
    prepend_text_codebase_url = MSG_CODEBASE_URL_DESCRIPTION

    if state.completed and state.vulns_count <= 0:
        text.append(MSG_NO_VULNERABILITIES_FOUND)
        prepend_text_codebase_url = MSG_NO_VULNS_CODEBASE_URL_DESCRIPTION
    else:
        for category in categories:
            category_count = getattr(state, category["count_attr"])

            if state.completed and category_count == 0:
                continue

            text.append(
                f"{category['icon']} {category['name']}: ", style=category["style"]
            )

            if category_count > 0:
                text.append(f"{category_count} ", style=category["style"])
                text.append(f"vulnerabilit{'y' if category_count == 1 else 'ies'}\n")
            else:
                text.append(
                    f"{spinner[(spinner_phase + category['spinner_offset']) % len(spinner)]} Scanning",
                    style=category["dim_style"],
                )
                text.append("\n")

        # Show fixes info if we have vulnerabilities
        if state.vulns_count > 0 and state.fixes is not None:
            text.append("\n")
            if state.fixes > 0:
                text.append("✨ ")
                text.append(f"{state.fixes} ", style="green")
                text.append(
                    f"automatic {'fix' if state.fixes == 1 else 'fixes'} available, resolving {state.fixed_vulns} vulnerabilities\n"
                )
            else:
                text.append(" No automatic fixes available for these vulnerabilities\n")

    # Dashboard link if URL is available
    if state.codebase_url is not None:
        text.append("\n")
        text.append(prepend_text_codebase_url)
        text.append(f"🔗 {state.codebase_url}\n", style="blue underline")
    elif state.completed:
        text.append("\n")

    return text


def generate_status_updates(state: InitScanState, spinner_phase=0):
    """
    Generate text displaying current status updates and progress information

    Args:
        state: The InitScanState object containing status information
        spinner_phase: Current phase of the spinner animation

    Returns:
        Rich Text object containing formatted status updates
    """
    text = Text()
    spinner = "⠋⠙⠹⠸⠼⠴⠦⠧⠇⠏"

    # Display current status message if available
    if state.status_message:
        action_symbol = {
            "init": "🔄",
            "scanning": "🔍",
            "uploading": "☁️ ",
            "complete": "✅",
            "error": "❌",
        }
        status_key = state.status_action if state.status_action is not None else "init"
        action_symbol = action_symbol.get(status_key, "ℹ️ ")

        text.append("\n")
        text.append(f"{action_symbol} Status: ", style="bold cyan")
        text.append(f"{state.status_message}\n", style="cyan")

        # If we're processing a file, show that
        if state.current_file and state.status_action == "scanning":
            text.append("📁 Current file: ", style="bold cyan")
            text.append(f"{state.current_file}\n", style="dim cyan")
            # Ensure progress is capped at 100%
            display_progress = min(state.progress, 100)
            text.append(f"📊 Progress: {display_progress}%\n", style="cyan")
    else:
        # If no status message, just show a spinner
        text.append(
            f"{spinner[spinner_phase % len(spinner)]} Scanning...", style="cyan"
        )

    return text


def process_scan_results(
    scan_results: Iterator["ScanResult"], state: InitScanState
) -> None:
    """Process the scan iterator and update state from typed result models

    Args:
        scan_results: Iterator yielding scan results from init_scan
        state: The InitScanState object to update with scan results
    """
    # Import the scan result types to handle typed results
    from safety.scan.init_scan import (
        ScanResultType,
    )

    try:
        for result in scan_results:
            # Now result is a typed model with proper attributes
            if result.type == ScanResultType.INIT:
                # Initial state with dependency count
                state.dependencies = result.dependencies
                state.status_message = "Initializing scan"
                state.status_action = "init"

            elif result.type == ScanResultType.PROGRESS:
                # Update all the state fields from the progress result
                # Type checker knows result is ProgressScanResult
                # Ensure progress never exceeds 100%
                state.progress = min(result.percent, 100)
                state.dependencies = result.dependencies

                # Track current file being processed
                state.current_file = result.file
                state.status_message = f"Processing {result.file}"
                state.status_action = "scanning"

                # Update severity counts if present
                if result.critical is not None:
                    state.critical = result.critical
                if result.high is not None:
                    state.high = result.high
                if result.medium is not None:
                    state.medium = result.medium
                if result.low is not None:
                    state.low = result.low
                if result.others is not None:
                    state.others = result.others

                # Update vulnerability count
                if result.vulns_count is not None:
                    state.vulns_count = result.vulns_count

                # Update fix information if present
                if result.fixes is not None:
                    state.fixes = result.fixes
                if result.fixed_vulns is not None:
                    state.fixed_vulns = result.fixed_vulns

            elif result.type == ScanResultType.STATUS:
                # Generic status update
                state.status_message = result.message
                state.status_action = result.action
                if result.percent is not None:
                    state.progress = min(result.percent, 100)

            elif result.type == ScanResultType.UPLOADING:
                # Status update for uploading phase
                state.status_message = result.message
                state.status_action = "uploading"
                if result.percent is not None:
                    state.progress = min(result.percent, 100)

            elif result.type == ScanResultType.COMPLETE:
                # Final update with complete data
                # Type checker knows result is CompleteScanResult
                state.progress = 100
                state.dependencies = result.dependencies
                state.critical = result.critical
                state.high = result.high
                state.medium = result.medium
                state.low = result.low
                state.others = result.others
                state.fixes = result.fixes
                state.fixed_vulns = result.fixed_vulns
                state.status_message = "Scan completed"
                state.status_action = "complete"
                state.vulns_count = result.vulns_count

                # Update project URL if available
                if result.codebase_url:
                    state.codebase_url = result.codebase_url

                if result.scan_id:
                    state.scan_id = result.scan_id

                # We're done processing
                state.completed = True

            # Add a small delay between updates to allow UI thread to refresh
            time.sleep(0.05)
    except Exception as e:
        console.print(f"Error processing scan results: {e}", style="bold red")
        state.status_message = f"Error: {str(e)}"
        state.status_action = "error"
    finally:
        # Ensure we mark as completed even if there was an exception
        state.completed = True


def init_scan_ui(ctx: "typer.Context", prompt_user: bool = False) -> InitScanState:
    """
    Initialize and run a scan for the init command, showing a live UI with scan progress.
    Uses the start_scan function to get an iterator of scan results and displays UI based on them.

    Args:
        ctx: The Typer context object containing configuration and project information
    """
    # Initialize state for tracking scan progress
    state = InitScanState()

    # Set up scan parameters and get the scan iterator
    target = ctx.obj.project.project_path.parent
    use_server_matching = False

    # Start the scan using the dedicated function
    scan_results = start_scan(
        ctx=ctx,
        target=target,
        use_server_matching=use_server_matching,
        auth_type=ctx.obj.auth.client.get_authentication_type(),
        is_authenticated=ctx.obj.auth.client.is_using_auth_credentials(),
        client=ctx.obj.auth.client,
        project=ctx.obj.project,
        platform_enabled=ctx.obj.platform_enabled,
    )

    # Process the scan results in a separate thread
    scan_thread = threading.Thread(
        target=process_scan_results, args=(scan_results, state)
    )
    scan_thread.daemon = True
    scan_thread.start()

    # Handle UI updates in the main thread
    spinner_phase = 0
    render_header(
        MSG_ANALYZE_CODEBASE_TITLE.format(project_name=ctx.obj.project.id), emoji="🔍"
    )
    time.sleep(0.8)

    # Detect if running on Windows
    is_windows = sys.platform == "win32"

    # Alternate screen in Windows is buggy, so we disable it
    live_kwargs = {
        "refresh_per_second": 10,
        "screen": False if is_windows else True,
        "transient": False,
    }

    refresh_sleep = 0.1

    with Live(**live_kwargs) as live:
        while not state.completed or scan_thread.is_alive():
            # Update spinner phase for animation
            spinner_phase = (spinner_phase + 1) % 10

            # Summary information shown below status updates
            summary = generate_summary(state, spinner_phase)

            if is_windows:
                content = summary
            else:
                # Create a container for all UI elements
                container = Table.grid(padding=0, expand=True)
                container.add_row(None)
                container.add_row(
                    Panel(
                        generate_status_updates(state, spinner_phase),
                        border_style="cyan",
                        padding=(0, 1),
                    )
                )
                container.add_row(None)
                container.add_row(summary)
                content = container

            # Display the updated UI
            live.update(content)

            time.sleep(refresh_sleep)

        # Last sync
        if state.completed:
            live.update(generate_summary(state, spinner_phase))
        time.sleep(2)

    # Windows is not using alternate screen, so summary is already rendered
    if not is_windows:
        # Final update to ensure completion state is shown
        console.print(generate_summary(state))

    if state.codebase_url:
        typed_print(MSG_OPEN_DASHBOARD_PROMPT, end_line=False)
        should_open = "y"

        if prompt_user:
            should_open = Prompt.ask(
                "",
                choices=["y", "n", "Y", "N"],
                default="y",
                show_default=False,
                show_choices=False,
                console=console,
            ).lower()

        if should_open == "y":
            typer.launch(state.codebase_url)

    return state


@init_app.command(
    cls=SafetyCLICommand,
    help=CODEBASE_INIT_HELP,
    name=CODEBASE_INIT_CMD_NAME,
    options_metavar="[OPTIONS]",
    context_settings={
        "allow_extra_args": True,
        "ignore_unknown_options": True,
        CONTEXT_COMMAND_TYPE: CommandType.BETA,
        CONTEXT_FEATURE_TYPE: FeatureType.FIREWALL,
    },
)
@handle_cmd_exception
@notify
def init(
    ctx: typer.Context,
    directory: Annotated[
        Path,
        typer.Argument(  # type: ignore
            exists=True,
            file_okay=False,
            dir_okay=True,
            writable=False,
            readable=True,
            resolve_path=True,
            show_default=False,
            help=CODEBASE_INIT_DIRECTORY_HELP,
        ),
    ] = Path("."),
):
    emit_init_started(ctx.obj.event_bus, ctx)
    # TODO: check if tty is available
    tracker = StepTracker()
    try:
        do_init(ctx, directory, tracker, prompt_user=console.is_interactive)
    except KeyboardInterrupt as e:
        emit_init_exited(ctx.obj.event_bus, ctx, exit_step=tracker.current_step)
        raise e


def do_init(
    ctx: typer.Context, directory: Path, tracker: StepTracker, prompt_user: bool = True
):
    """
    Initialize Safety CLI with the new onboarding flow.

    Args:
        ctx: The CLI context
        directory: The target directory to initialize
        prompt_user: Whether to prompt the user for input or use defaults
    """
    project_dir = directory.resolve()

    typed_print(MSG_WELCOME_TITLE)
    progressive_print(MSG_WELCOME_DESCRIPTION)

    tracker.current_step = InitExitStep.PRE_AUTH
    org_slug = launch_auth_if_needed(ctx, console)
    tracker.current_step = InitExitStep.POST_AUTH

    render_header(MSG_SETUP_PACKAGE_FIREWALL_TITLE, margin_right=1)
    console.print(MSG_SETUP_PACKAGE_FIREWALL_DESCRIPTION)

    console.print(
        Syntax(
            MSG_FIREWALL_UNINSTALL, "bash", theme="monokai", background_color="default"
        )
    )

    completed_tools = ""
    all_completed = False
    all_missing = True

    status = {
        ToolType.PIP: {
            "alias": AliasConfig(is_configured=False),
            "index": IndexConfig(is_configured=False),
        },
        ToolType.POETRY: {
            "alias": AliasConfig(is_configured=False),
            "index": IndexConfig(is_configured=False),
        },
        ToolType.UV: {
            "alias": AliasConfig(is_configured=False),
            "index": IndexConfig(is_configured=False),
        },
    }

    tracker.current_step = InitExitStep.PRE_FIREWALL_SETUP
    if ask_firewall_setup(ctx, prompt_user):
        completed_tools, all_completed, all_missing, status = setup_firewall(
            ctx, status, org_slug, console
        )
        console.line()
        ask_continue(ctx, prompt_user)
        console.line()

    tracker.current_step = InitExitStep.POST_FIREWALL_SETUP

    render_header(MSG_SETUP_CODEBASE_TITLE, emoji="🔒")
    console.print(MSG_SETUP_CODEBASE_DESCRIPTION)

    project_scan_state = None

    tracker.current_step = InitExitStep.PRE_CODEBASE_SETUP
    local_files = find_local_tool_files(project_dir)

    emit_codebase_detection_status(
        event_bus=ctx.obj.event_bus,
        ctx=ctx,
        detected=any(local_files),
        detected_files=local_files if local_files else None,
    )

    if local_files:
        progressive_print(
            [
                f"{load_emoji('📌')} We found a `{file.name}` file in this directory."
                for file in local_files
            ]
        )

        console.line()

        if ask_codebase_setup(ctx, prompt_user):
            project_created, project_status = create_project(ctx, console, project_dir)

            configure_local_directory(project_dir, org_slug, ctx.obj.project.id)

            emit_codebase_setup_completed(
                event_bus=ctx.obj.event_bus,
                ctx=ctx,
                is_created=project_created,
                codebase_id=ctx.obj.project.id if project_created else None,
            )

            if project_created:
                console.print(
                    "\n" + f"{ctx.obj.project.id} codebase {project_status} ✅"
                )
            else:
                progressive_print([f"{load_emoji('x')} Failed to create codebase"])

            console.line()

            tracker.current_step = InitExitStep.PRE_SCAN
            project_scan_state = init_scan_ui(ctx, prompt_user)
            tracker.current_step = InitExitStep.POST_SCAN
            emit_init_scan_completed(
                event_bus=ctx.obj.event_bus,
                ctx=ctx,
                scan_id=project_scan_state.scan_id,
            )
    else:
        console.print(MSG_SETUP_CODEBASE_NO_PROJECT)

    tracker.current_step = InitExitStep.POST_CODEBASE_SETUP

    console.line()
    render_header(MSG_SETUP_COMPLETE_TITLE, emoji="🏆")

    is_setup_complete = all_completed and project_scan_state

    wrap_up_msg = []

    if all_completed:
        wrap_up_msg.append(
            MSG_COMPLETE_TOOL_SECURED.format(
                tools=completed_tools,
                firewall_url="https://platform.safetycli.com/firewall/",
            )
        )
    elif all_missing:
        wrap_up_msg.append(Text.from_markup(MSG_TOOLS_NOT_CONFIGURED))
    else:
        wrap_up_msg.append(Text.from_markup(MSG_SETUP_INCOMPLETE))

    wrap_up_msg.append("")

    if project_scan_state:
        if project_scan_state.scan_id:
            wrap_up_msg.append(
                MSG_COMPLETE_SECURED.format(
                    codebase_url=project_scan_state.codebase_url
                )
            )
        else:
            msg = (
                project_scan_state.status_message
                if project_scan_state.status_message
                else "Unknown"
            )
            wrap_up_msg.append(
                Text.from_markup(MSG_CODEBASE_FAILED_TO_SCAN.format(reason=msg))
            )
    else:
        wrap_up_msg.append(Text.from_markup(MSG_CODEBASE_NOT_CONFIGURED))

    if wrap_up_msg:
        progressive_print(wrap_up_msg)
        console.line()

        if is_setup_complete:
            typed_print(MSG_SETUP_COMPLETE_SUBTITLE)
            console.line()
            typed_print(MSG_LAST_MANUAL_STEP)
            console.line()

    render_header(title=MSG_SETUP_NEXT_STEPS_SUBTITLE, emoji="🚀")
    console.line()

    next_steps_msg = MSG_SETUP_NEXT_STEPS

    if not project_scan_state:
        next_steps_msg = MSG_SETUP_NEXT_STEPS_NO_PROJECT
    elif project_scan_state.vulns_count <= 0:
        next_steps_msg = MSG_SETUP_NEXT_STEPS_NO_VULNS

    progressive_print(
        [Padding(Text.from_markup(line), (0, 0, 1, 0)) for line in next_steps_msg]
    )

    console.line()
    typed_print(MSG_SETUP_NEXT_STEPS_MANUAL_STEP, delay=0.04)
    console.line()

    # Emit event for firewall configuration
    emit_firewall_configured(
        event_bus=ctx.obj.event_bus,
        status=status,
    )
    tracker.current_step = InitExitStep.COMPLETED



import logging
from typing import List, Dict, Tuple, Optional


from safety.formatter import FormatterAPI
from safety.formatters.json import build_json_report
from safety.output_utils import get_report_brief_info, parse_html
from safety.util import get_basic_announcements

LOG = logging.getLogger(__name__)


class HTMLReport(FormatterAPI):
    """
    HTML report formatter for when the output is input for something else.
    """

    def render_vulnerabilities(self, announcements: List[Dict], vulnerabilities: List[Dict], remediations: Dict,
                               full: bool, packages: List[Dict], fixes: Tuple = ()) -> Optional[str]:
        """
        Renders vulnerabilities in HTML format.

        Args:
            announcements (List[Dict]): List of announcements.
            vulnerabilities (List[Dict]): List of vulnerabilities.
            remediations (Dict): Remediation data.
            full (bool): Flag indicating full output.
            packages (List[Dict]): List of packages.
            fixes (Tuple, optional): Tuple of fixes.

        Returns:
            str: Rendered HTML vulnerabilities report.
        """
        LOG.debug(
            f'HTML Output, Rendering {len(vulnerabilities)} vulnerabilities, {len(remediations)} package '
            f'remediations with full_report: {full}')
        report = build_json_report(announcements, vulnerabilities, remediations, packages)

        return parse_html(kwargs={"json_data": report})

    def render_licenses(self, announcements: List[Dict], licenses: List[Dict]) -> None:
        """
        Renders licenses in HTML format.

        Args:
            announcements (List[Dict]): List of announcements.
            licenses (List[Dict]): List of licenses.
        """
        pass

    def render_announcements(self, announcements: List[Dict]) -> None:
        """
        Renders announcements in HTML format.

        Args:
            announcements (List[Dict]): List of announcements.
        """
        pass



from importlib.metadata import PackageNotFoundError, version
import logging
import os
from typing import Dict, Optional


LOG = logging.getLogger(__name__)


def get_version() -> Optional[str]:
    """
    Get the version of the Safety package.

    Returns:
      Optional[str]: The Safety version if found, otherwise None.
    """
    try:
        return version("safety")
    except PackageNotFoundError:
        LOG.exception("Unable to get Safety version.")
        return None


def get_identifier() -> str:
    """
    Get the identifier of the source type.

    Returns:
      str: The source type identifier.
    """

    if source := os.environ.get("SAFETY_SOURCE_TYPE", None):
        return source

    from safety_schemas.models.events.types import SourceType

    return SourceType.SAFETY_CLI_PYPI.value


def get_meta_http_headers() -> Dict[str, str]:
    """
    Get the metadata headers for the client.

    Returns:
      Dict[str, str]: The metadata headers.
    """

    from safety_schemas.models.events.constants import SAFETY_NAMESPACE

    namespace = SAFETY_NAMESPACE.title()

    return {
        f"{namespace}-Client-Version": get_version() or "",
        f"{namespace}-Client-Id": get_identifier(),
    }



import os
import unittest
from datetime import datetime
from unittest.mock import Mock, patch

from packaging.version import parse

from safety.models import Package, SafetyRequirement
from safety.output_utils import format_vulnerability, get_printable_list_of_scanned_items, build_remediation_section, \
    get_final_brief_license, get_fix_hint_for_unpinned
from tests.test_cli import get_vulnerability


class TestOutputUtils(unittest.TestCase):

    def setUp(self) -> None:
        self.maxDiff = None

    @patch("safety.output_utils.is_using_api_key")
    def test_format_vulnerability(self, is_using_api_key):
        is_using_api_key.return_value = True

        numpy_pkg = {'name': 'numpy', 'version': '1.22.0', 'requirements': [SafetyRequirement('numpy==1.22.0')],
                     'secure_versions': ['1.22.3'],
                     'insecure_versions': ['1.22.2', '1.22.1', '1.22.0', '1.22.0rc3', '1.21.5']}
        severity = {
            "cvssv2": {
                "base_score": 4.3,
                "impact_score": 2.9,
                "vector_string": "AV:N/AC:M/Au:N/C:N/I:P/A:N"
            },
            "cvssv3": {
                "base_score": 6.1,
                "impact_score": 2.7,
                "base_severity": "MEDIUM",
                "vector_string": "CVSS:3.1/AV:N/AC:L/PR:N/UI:R/S:C/C:L/I:L/A:N"
            }}

        vulnerability = get_vulnerability(pkg_kwargs=numpy_pkg,
                                          vuln_kwargs={'affected_versions': ['1.22.0', '1.21.5']},
                                          cve_kwargs=severity)
        columns = 80

        output = format_vulnerability(vulnerability, full_mode=True, only_text=True, columns=columns)

        lines = [
            '-> Vulnerability found in numpy version 1.22.0',
            '   Vulnerability ID: PYUP-1234',
            '   Affected spec: >0',
            '   ADVISORY: ',
            '   Fixed versions: No known fix',
            '   bla is MEDIUM SEVERITY => CVSS v3, BASE SCORE 6.1, IMPACT',
            '   SCORE 2.7, VECTOR STRING CVSS:3.1/AV:N/AC:L/PR:N/UI:R/S:C/C:L/I:L/A:N',
            '   CVSS v2, BASE SCORE 4.3, IMPACT SCORE 2.9, VECTOR STRING',
            '   AV:N/AC:M/Au:N/C:N/I:P/A:N',
            '   For more information about this vulnerability, visit',
            '   https://pyup.io/PVE/2323',
            '   To ignore this vulnerability, use PyUp vulnerability id PYUP-1234 in',
            '   safety’s ignore command-line argument or add the ignore to your safety',
            '   policy file.\n'
        ]

        EXPECTED = '\n'.join(lines)
        self.assertEqual(output, EXPECTED)

    @patch("safety.output_utils.is_using_api_key")
    def test_format_vulnerability_with_dep_spec(self, is_using_api_key):
        is_using_api_key.return_value = True

        numpy_pkg = {'name': 'numpy', 'version': '1.22.0', 'requirements': [SafetyRequirement('numpy>=1.22.0')],
                     'secure_versions': ['1.22.3'], 'insecure_versions': ['1.22.2', '1.22.1', '1.22.0', '1.22.0rc3',
                                                                          '1.21.5']}
        severity = {
            "cvssv2": {
                "base_score": 4.3,
                "impact_score": 2.9,
                "vector_string": "AV:N/AC:M/Au:N/C:N/I:P/A:N"
            },
            "cvssv3": {
                "base_score": 6.1,
                "impact_score": 2.7,
                "base_severity": "MEDIUM",
                "vector_string": "CVSS:3.1/AV:N/AC:L/PR:N/UI:R/S:C/C:L/I:L/A:N"
            }}

        vulnerability = get_vulnerability(pkg_kwargs=numpy_pkg,
                                          vuln_kwargs={'affected_versions': ['1.22.0', '1.21.5']},
                                          cve_kwargs=severity)
        columns = 80

        output = format_vulnerability(vulnerability, full_mode=True, only_text=True, columns=columns)

        # As we apply the style first and then we wrap the text, it won't use all the 80 columns.
        lines = [
            '-> Vulnerability may be present given that your numpy install specifier is',
            '   >=1.22.0',
            '   Vulnerability ID: PYUP-1234',
            '   Affected spec: >0',
            '   ADVISORY: ',
            '   Fixed versions: No known fix',
            '   bla is MEDIUM SEVERITY => CVSS v3, BASE SCORE 6.1, IMPACT',
            '   SCORE 2.7, VECTOR STRING CVSS:3.1/AV:N/AC:L/PR:N/UI:R/S:C/C:L/I:L/A:N',
            '   CVSS v2, BASE SCORE 4.3, IMPACT SCORE 2.9, VECTOR STRING',
            '   AV:N/AC:M/Au:N/C:N/I:P/A:N',
            '   This vulnerability is present in your install specifier range.',
            '   To learn more about reporting these, specifier range handling, and',
            '   options for scanning unpinned packages visit',
            '   https://docs.pyup.io/docs/safety-range-specs',
            '   For more information about this vulnerability, visit',
            '   https://pyup.io/PVE/2323',
            '   To ignore this vulnerability, use PyUp vulnerability id PYUP-1234 in',
            '   safety’s ignore command-line argument or add the ignore to your safety',
            '   policy file.\n'
        ]

        EXPECTED = '\n'.join(lines)

        self.assertEqual(output, EXPECTED)

    @patch("safety.output_utils.is_using_api_key")
    def test_format_vulnerability_with_ignored_vulnerability(self, is_using_api_key):
        is_using_api_key.return_value = True

        numpy_pkg = {'name': 'numpy', 'version': '1.22.0', 'requirements': [SafetyRequirement('numpy==1.22.0')],
                     'secure_versions': ['1.22.3'],
                     'insecure_versions': ['1.22.2', '1.22.1', '1.22.0', '1.22.0rc3', '1.21.5']}

        vulnerability = get_vulnerability(pkg_kwargs=numpy_pkg,
                                          vuln_kwargs={'ignored': True})
        columns = 80

        output = format_vulnerability(vulnerability, full_mode=True, only_text=True, columns=columns)

        lines = [
            '-> Vulnerability found in numpy version 1.22.0',
            '   Vulnerability ID: PYUP-1234',
            '   This vulnerability is being ignored.',
            '   For more information about this vulnerability, visit',
            '   https://pyup.io/PVE/2323\n'
        ]

        EXPECTED = '\n'.join(lines)
        self.assertEqual(output, EXPECTED)

        reason = 'We do not use that feature'
        vulnerability = get_vulnerability(pkg_kwargs=numpy_pkg,
                                          vuln_kwargs={'ignored': True,
                                                       'ignored_reason': reason})

        output = format_vulnerability(vulnerability, full_mode=True, only_text=True, columns=columns)

        lines = [
            '-> Vulnerability found in numpy version 1.22.0',
            '   Vulnerability ID: PYUP-1234',
            '   This vulnerability is being ignored.',
           f'   Reason: {reason}',
            '   For more information about this vulnerability, visit',
            '   https://pyup.io/PVE/2323\n',
        ]

        EXPECTED = '\n'.join(lines)
        self.assertEqual(output, EXPECTED)

        reason = 'We do not think this vuln is a security issue'
        expires = datetime.strptime('2022-06-21', '%Y-%m-%d')
        vulnerability = get_vulnerability(pkg_kwargs=numpy_pkg,
                                          vuln_kwargs={'ignored': True,
                                                       'ignored_reason': reason,
                                                       'ignored_expires': expires})

        expires = '2022-06-21 00:00:00 UTC'
        output = format_vulnerability(vulnerability, full_mode=True, only_text=True, columns=120)

        lines = [
            '-> Vulnerability found in numpy version 1.22.0',
            '   Vulnerability ID: PYUP-1234',
           f'   This vulnerability is being ignored until {expires}. See your configurations.',
           f'   Reason: {reason}',
            '   For more information about this vulnerability, visit https://pyup.io/PVE/2323\n'
        ]

        EXPECTED = '\n'.join(lines)
        self.assertEqual(output, EXPECTED)

    @patch("safety.output_utils.SafetyContext")
    def test_get_printable_list_of_scanned_items_stdin(self, ctx):
        ctx.return_value = Mock(packages=[])
        output = get_printable_list_of_scanned_items('stdin')

        EXPECTED = (
            [[{'styled': False, 'value': 'No found packages in stdin'}]],
            ['No found packages in stdin'])

        self.assertTupleEqual(output, EXPECTED)

        p_kwargs = {'name': 'django', 'version': '2.2', 'requirements': [SafetyRequirement('django==2.2')],
                    'found': '/site-packages/django',
                    'insecure_versions': [], 'secure_versions': ['2.2'],
                    'latest_version_without_known_vulnerabilities': '2.2', 'latest_version': '2.2',
                    'more_info_url': 'https://pyup.io/package/foo'}
        ctx.return_value = Mock(packages=[Package(**p_kwargs)])
        output = get_printable_list_of_scanned_items('stdin')

        EXPECTED = (
            [[{'styled': False, 'value': 'django'}]],
            ['django'])

        self.assertTupleEqual(output, EXPECTED)

    @patch("safety.output_utils.SafetyContext")
    def test_get_printable_list_of_scanned_items_environment(self, ctx):
        ctx.return_value = Mock(packages=[], scanned_full_path=[])
        output = get_printable_list_of_scanned_items('environment')

        no_locations = 'No locations found in the environment'

        EXPECTED = (
            [[{'styled': False, 'value': no_locations}]],
            [no_locations])

        self.assertTupleEqual(output, EXPECTED)

    @patch("safety.output_utils.SafetyContext")
    def test_get_printable_list_of_scanned_items_files(self, ctx):
        dirname = os.path.dirname(__file__)
        file_a = open(os.path.join(dirname, "reqs_1.txt"), mode='r')
        file_b = open(os.path.join(dirname, "reqs_4.txt"), mode='r')

        ctx.return_value = Mock(params={'files': [file_a, file_b]})
        output = get_printable_list_of_scanned_items('files')

        EXPECTED = ([
                        [{'styled': False, 'value': f'-> {file_a.name}'}],
                        [{'styled': False, 'value': f'-> {file_b.name}'}]
                    ],
                    [file_a.name, file_b.name])

        self.assertTupleEqual(output, EXPECTED)

    @patch("safety.output_utils.SafetyContext")
    def test_get_printable_list_of_scanned_items_file(self, ctx):
        # Used by the review command
        report = open(os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "test_db",
            "report.json"
        ), mode='r')

        ctx.return_value = Mock(params={'file': report})
        output = get_printable_list_of_scanned_items('file')

        EXPECTED = (
            [[{'styled': False, 'value': f'-> {report.name}'}]],
            [report.name])

        self.assertTupleEqual(output, EXPECTED)

    @patch("safety.output_utils.is_using_api_key")
    def test_build_remediation_section_with_api_key(self, is_using_api_key):
        is_using_api_key.return_value = True

        remediations = {
            'django': {
                '==4.0.1': {'vulnerabilities_found': 1, 'version': '4.0.1',
                            'requirement': SafetyRequirement('django==4.0.1'),
                            'other_recommended_versions': ['2.2.28', '3.2.13'],
                            'recommended_version': parse('4.0.4'),
                            'closest_secure_version': {'upper': parse('4.0.4'), 'lower': None},
                            'more_info_url': 'https://pyup.io/packages/pypi/django/?from=4.0.1&to=4.0.4'}}
        }

        EXPECTED = ['   REMEDIATIONS',
                    '\n-> django version 4.0.1 was found, which has 1 vulnerability'
                    '\n   The closest version with no known vulnerabilities is 4.0.4'
                    '\n' + f"{' ' * 78}"
                    '\n   We recommend updating to version 4.0.4 of django. Other versions'
                    '\n   without known vulnerabilities are: 2.2.28, 3.2.13'
                    '\n' + f"{' ' * 78}"
                    '\n   For more information about the django package and update options, visit'
                    '\n   https://pyup.io/packages/pypi/django/?from=4.0.1&to=4.0.4'
                    '\n   Always check for breaking changes when updating packages.'
                    '\n' + f"{' ' * 78}",
                    '+==============================================================================+']

        rem_text_section = build_remediation_section(remediations=remediations, only_text=True, columns=80)

        self.assertEqual(rem_text_section, EXPECTED)

    @patch("safety.output_utils.is_using_api_key")
    def test_build_remediation_section_without_api_key(self, is_using_api_key):
        is_using_api_key.return_value = False

        remediations = {
            'django': {
                '==4.0.1': {'vulnerabilities_found': 1, 'version': '4.0.1',
                            'requirement': SafetyRequirement('django==4.0.1'),
                            'secure_versions': ['2.2.28', '3.2.13', '4.0.4'],
                            'closest_secure_version': {'major': parse('4.0.4'), 'minor': None},
                            'more_info_url': 'https://pyup.io/packages/pypi/django/'}}}

        # Start & End line decorator in format_long_text affects this output
        EXPECTED = ['   REMEDIATIONS',
                    '\n  1 vulnerability was reported in 1 package. For detailed remediation & fix '
                    '\n  recommendations, upgrade to a commercial license. \n',
                    '+==============================================================================+']

        rem_text_section = build_remediation_section(remediations=remediations, only_text=True, columns=80)

        self.assertEqual(rem_text_section, EXPECTED)

    def test_get_final_brief_license(self):
        licenses = set(["MIT"])
        EXPECTED = f' The following software licenses were present in your system: {", ".join(licenses)} '
        brief = get_final_brief_license(licenses)
        self.assertEqual(EXPECTED, brief)

        licenses = set([])
        EXPECTED = '  Scan was completed. '
        brief = get_final_brief_license(licenses)
        self.assertEqual(EXPECTED, brief)

    def test_get_fix_hint_for_unpinned_no_other_recommended_versions(self):
        req_rem = {
            'vulnerabilities_found': 2,
            'version': None,
            'requirement': SafetyRequirement('django>=2.2;python_version>="3.6"'),
            'more_info_url': 'https://pyup.io/django',
            'recommended_version': '3.2.18',
            'other_recommended_versions': [],
            'closest_secure_version': {
                'upper': '3.2.18',
                'lower': None
            }
        }

        result = get_fix_hint_for_unpinned(req_rem)
        self.assertEqual(
            "Version 3.2.18 has no known vulnerabilities and falls within your current specifier range.",
            result
        )

    def test_get_fix_hint_for_unpinned_with_other_recommended_versions(self):
        req_rem = {
            'vulnerabilities_found': 2,
            'version': None,
            'requirement': SafetyRequirement('django>=2.2'),
            'more_info_url': 'https://pyup.io/django',
            'recommended_version': '3.2.20',
            'other_recommended_versions': ['4.1.7', '4.2b1'],
            'closest_secure_version': {
                'upper': '3.2.18',
                'lower': None
            }
        }

        result = get_fix_hint_for_unpinned(req_rem)
        self.assertEqual(
            "Version 3.2.20 has no known vulnerabilities and falls within your current specifier range."
            " Other versions without known vulnerabilities are: 4.1.7, 4.2b1",
            result
        )

    def test_get_fix_hint_for_unpinned_with_one_other_recommended_version(self):
        req_rem = {
            'vulnerabilities_found': 2,
            'version': None,
            'requirement': SafetyRequirement('django>=2.2'),
            'more_info_url': 'https://pyup.io/django',
            'recommended_version': '3.2.20',
            'other_recommended_versions': ['4.1.7'],
            'closest_secure_version': {
                'upper': '3.2.18',
                'lower': None
            }
        }

        result = get_fix_hint_for_unpinned(req_rem)
        self.assertEqual(
            "Version 3.2.20 has no known vulnerabilities and falls within your current specifier range."
            " Other version without known vulnerabilities is 4.1.7",
            result
        )



from functools import wraps
import logging
import os
from pathlib import Path
from typing import Any, List, Optional

from rich.padding import Padding
from safety_schemas.models import ConfigModel, ProjectModel
from rich.console import Console
from safety.auth.cli import render_email_note
from safety.cli_util import process_auth_status_not_ready
from safety.console import main_console
from safety.constants import SYSTEM_POLICY_FILE, USER_POLICY_FILE
from safety.errors import SafetyError, SafetyException
from safety.scan.constants import DEFAULT_SPINNER
from safety.scan.main import download_policy, load_policy_file, resolve_policy
from safety.scan.models import ScanOutput, SystemScanOutput
from safety.scan.render import print_announcements, print_header, print_wait_policy_download
from safety.scan.util import GIT
from ..init.main import load_unverified_project_from_config, verify_project

from safety.scan.validators import fail_if_not_allowed_stage

from safety.util import build_telemetry_data, pluralize
from safety_schemas.models import MetadataModel, ScanType, ReportSchemaVersion, \
    PolicySource

LOG = logging.getLogger(__name__)


def scan_project_command_init(func):
    """
    Decorator to make general verifications before each project scan command.
    """
    @wraps(func)
    def inner(ctx, policy_file_path: Optional[Path], target: Path,
              output: ScanOutput,
              console: Console = main_console,
              *args, **kwargs):
        ctx.obj.console = console
        ctx.params.pop("console", None)

        if output.is_silent():
            console.quiet = True

        if not ctx.obj.auth.is_valid():
            process_auth_status_not_ready(console=console,
                                        auth=ctx.obj.auth, ctx=ctx)

        upload_request_id = kwargs.pop("upload_request_id", None)

        # Load .safety-project.ini
        unverified_project = load_unverified_project_from_config(project_root=target)

        print_header(console=console, targets=[target])

        stage = ctx.obj.auth.stage
        session = ctx.obj.auth.client
        git_data = GIT(root=target).build_git_data()
        origin = None
        branch = None

        if git_data:
            origin = git_data.origin
            branch = git_data.branch

        if ctx.obj.platform_enabled:
            verify_project(console, ctx, session, unverified_project, stage, origin)
        else:
            ctx.obj.project = ProjectModel(
                id="",
                name="Undefined project",
                project_path=unverified_project.project_path
            )

        ctx.obj.project.git = git_data
        ctx.obj.project.upload_request_id = upload_request_id

        if not policy_file_path:
            policy_file_path = target / Path(".safety-policy.yml")

        # Load Policy file and pull it from CLOUD
        local_policy = kwargs.pop("local_policy",
                                  load_policy_file(policy_file_path))

        cloud_policy = None
        if ctx.obj.platform_enabled:
            cloud_policy = print_wait_policy_download(console, (download_policy,
                                            {"session": session,
                                            "project_id": ctx.obj.project.id,
                                            "stage": stage,
                                            "branch": branch}))

        ctx.obj.project.policy = resolve_policy(local_policy, cloud_policy)
        config = ctx.obj.project.policy.config \
            if ctx.obj.project.policy and ctx.obj.project.policy.config \
                else ConfigModel()

        # Preserve global telemetry preference.
        if ctx.obj.config:
            if ctx.obj.config.telemetry_enabled is not None:
                config.telemetry_enabled = ctx.obj.config.telemetry_enabled

        ctx.obj.config = config

        console.print()

        if ctx.obj.auth.org and ctx.obj.auth.org.name:
            console.print(f"[bold]Organization[/bold]: {ctx.obj.auth.org.name}")

        # Check if an API key is set
        if ctx.obj.auth.client.get_authentication_type() == "api_key":
            details = {"Account": f"API key used"}
        else:

            if ctx.obj.auth.client.get_authentication_type() == "token":
                content = ctx.obj.auth.email
                if ctx.obj.auth.name != ctx.obj.auth.email:
                    content = f"{ctx.obj.auth.name}, {ctx.obj.auth.email}"

                details = {"Account": f"{content} {render_email_note(ctx.obj.auth)}"}
            else:
                details = {"Account": f"Offline - {os.getenv('SAFETY_DB_DIR')}"}

        if ctx.obj.project.id:
            details["Project"] = ctx.obj.project.id

        if ctx.obj.project.git:
            details[" Git branch"] = ctx.obj.project.git.branch

        details[" Environment"] = ctx.obj.auth.stage

        msg = "None, using Safety CLI default policies"

        if ctx.obj.project.policy:
            if ctx.obj.project.policy.source is PolicySource.cloud:
                msg = f"fetched from Safety Platform, " \
                    "ignoring any local Safety CLI policy files"
            else:
                if ctx.obj.project.id:
                    msg = f"local {ctx.obj.project.id} project scan policy"
                else:
                    msg = f"local scan policy file"

        details[" Scan policy"] = msg

        for k,v in details.items():
            console.print(f"[scan_meta_title]{k}[/scan_meta_title]: {v}")

        print_announcements(console=console, ctx=ctx)

        console.print()

        result = func(ctx, target=target, output=output, *args, **kwargs)


        return result

    return inner


def scan_system_command_init(func):
    """
    Decorator to make general verifications before each system scan command.
    """
    @wraps(func)
    def inner(ctx, policy_file_path: Optional[Path], targets: List[Path],
              output: SystemScanOutput,
              console: Console = main_console, *args, **kwargs):
        ctx.obj.console = console
        ctx.params.pop("console", None)

        if output.is_silent():
            console.quiet = True

        if not ctx.obj.auth.is_valid():
            process_auth_status_not_ready(console=console,
                                        auth=ctx.obj.auth, ctx=ctx)

        console.print()
        print_header(console=console, targets=targets, is_system_scan=True)

        if not policy_file_path:
            if SYSTEM_POLICY_FILE.exists():
                policy_file_path = SYSTEM_POLICY_FILE
            elif USER_POLICY_FILE.exists():
                policy_file_path = USER_POLICY_FILE

        # Load Policy file
        ctx.obj.system_scan_policy = load_policy_file(policy_file_path) if policy_file_path else None
        config = ctx.obj.system_scan_policy.config \
            if ctx.obj.system_scan_policy and ctx.obj.system_scan_policy.config \
                else ConfigModel()

        # Preserve global telemetry preference.
        if ctx.obj.config:
            if ctx.obj.config.telemetry_enabled is not None:
                config.telemetry_enabled = ctx.obj.config.telemetry_enabled

        ctx.obj.config = config

        if not any(targets):
            if any(config.scan.system_targets):
                targets = [Path(t).expanduser().absolute() for t in config.scan.system_targets]
            else:
                targets = [Path("/")]

            ctx.obj.metadata.scan_locations = targets

        console.print()

        if ctx.obj.auth.org and ctx.obj.auth.org.name:
            console.print(f"[bold]Organization[/bold]: {ctx.obj.auth.org.name}")

        details = {"Account": f"{ctx.obj.auth.name}, {ctx.obj.auth.email}",
                "Scan stage": ctx.obj.auth.stage}

        if ctx.obj.system_scan_policy:
            if ctx.obj.system_scan_policy.source is PolicySource.cloud:
                policy_type = "remote"
            else:
                policy_type = f'local ("{ctx.obj.system_scan_policy.id}")'

            org_name = " "
            if ctx.obj.auth.org and ctx.obj.auth.org.name:
                org_name = f" {ctx.obj.auth.org.name} "

            details["System scan policy"] = f"{policy_type}{org_name}organization policy:"

        for k,v in details.items():
            console.print(f"[bold]{k}[/bold]: {v}")

        if ctx.obj.system_scan_policy:

            dirs = [ign for ign in ctx.obj.config.scan.ignore if Path(ign).is_dir()]

            policy_details = [
                f"-> scanning from root {', '.join([str(t) for t in targets])} to a max folder depth of {ctx.obj.config.scan.max_depth}",
                f"-> excluding {len(dirs)} {pluralize('directory', len(dirs))} and their sub-directories",
                "-> target ecosystems: Python"
            ]
            for policy_detail in policy_details:
                console.print(
                    Padding(policy_detail,
                            (0, 0, 0, 1)), emoji=True)

        print_announcements(console=console, ctx=ctx)

        console.print()

        kwargs.update({"targets": targets})
        result = func(ctx, *args, **kwargs)
        return result

    return inner


def inject_metadata(func):
    """
    Build metadata per subcommand. A system scan can trigger a project scan,
    the project scan will need to build its own metadata.
    """
    @wraps(func)
    def inner(ctx, *args, **kwargs):
        telemetry = build_telemetry_data(telemetry=ctx.obj.config.telemetry_enabled,
                                         command=ctx.command.name,
                                         subcommand=ctx.invoked_subcommand)

        auth_type = ctx.obj.auth.client.get_authentication_type()

        scan_type = ScanType(ctx.command.name)
        target = kwargs.get("target", None)
        targets = kwargs.get("targets", None)

        if not scan_type:
            raise SafetyException("Missing scan_type.")

        if scan_type is ScanType.scan:
            if not target:
                raise SafetyException("Missing target.")
            targets = [target]

        metadata = MetadataModel(
            scan_type=scan_type,
            stage=ctx.obj.auth.stage,
            scan_locations=targets,
            authenticated=ctx.obj.auth.client.is_using_auth_credentials(),
            authentication_type=auth_type,
            telemetry=telemetry,
            schema_version=ReportSchemaVersion.v3_0
            )

        ctx.obj.schema = ReportSchemaVersion.v3_0
        ctx.obj.metadata = metadata
        ctx.obj.telemetry = telemetry

        return func(ctx, *args, **kwargs)

    return inner



from typing import Dict, List, Optional

from ..base import ToolCommandLineParser, CommandToolIntention
from ..intents import Dependency, ToolIntentionType


class PoetryParser(ToolCommandLineParser):
    def get_tool_name(self) -> str:
        return "poetry"

    @property
    def intention_mapping(self) -> Dict[str, ToolIntentionType]:
        """
        Maps specific commands to intention types
        """
        return {
            "add": ToolIntentionType.ADD_PACKAGE,
            "remove": ToolIntentionType.REMOVE_PACKAGE,
            "update": ToolIntentionType.UPDATE_PACKAGE,
            "show": ToolIntentionType.LIST_PACKAGES,
            "init": ToolIntentionType.INIT_PROJECT,
            "build": ToolIntentionType.BUILD_PROJECT,
            "install": ToolIntentionType.SYNC_PACKAGES,
        }

    def parse(self, args: List[str]) -> Optional[CommandToolIntention]:
        """
        Parse the command line arguments into a CommandToolIntention object.

        Args:
            args (List[str]): Command line arguments

        Returns:
            Optional[CommandToolIntention]: Parsed command tool intention
        """
        if not self.can_handle(args):
            return None

        command = args[0].lower()
        intention_type = self.map_intention(command)
        options = {}
        packages = []
        raw_args = args.copy()

        # Process command arguments based on intention type
        i = 1  # Skip command
        while i < len(args):
            arg = args[i]

            if arg.startswith("-"):
                # Handle options
                option_key = arg.lstrip("-")

                # Handle option with value
                if i + 1 < len(args) and not args[i + 1].startswith("-"):
                    option_value = args[i + 1]
                    options[option_key] = {
                        "arg_index": i,
                        "raw_option": arg,
                        "value": option_value,
                        "value_index": i + 1,
                    }
                    i += 2
                else:
                    # Flag option (no value)
                    options[option_key] = {
                        "arg_index": i,
                        "raw_option": arg,
                        "value": True,
                    }
                    i += 1
            else:
                # Parse non-option arguments
                if intention_type == ToolIntentionType.ADD_PACKAGE:
                    try:
                        # Parse package specification
                        dep = self._parse_package_spec(arg, i)
                        if not dep:
                            # Let's skip this dependency
                            i += 1
                            continue

                        # Check if this is a dev dependency
                        is_dev = any(opt in ["dev", "D"] for opt in options.keys())
                        dep.is_dev_dependency = is_dev

                        packages.append(dep)
                    except ValueError:
                        # Not a valid package spec, store as unknown option
                        options[f"unknown_{len(options)}"] = {
                            "arg_index": i,
                            "value": arg,
                        }
                elif intention_type in [
                    ToolIntentionType.REMOVE_PACKAGE,
                    ToolIntentionType.UPDATE_PACKAGE,
                ]:
                    # Handle packages for remove/update commands
                    # These often have simpler formats than add commands
                    dep = Dependency(name=arg, arg_index=i, original_text=arg)
                    packages.append(dep)
                i += 1

        return CommandToolIntention(
            tool=self._tool_name,
            command=command,
            intention_type=intention_type,
            packages=packages,
            options=options,
            raw_args=raw_args,
        )

    def _parse_package_spec(
        self, spec_str: str, arg_index: int
    ) -> Optional[Dependency]:
        """
        Parse a package specification string into a Dependency object.
        Handles various formats including Poetry-specific syntax and standard PEP 508 requirements.

        Args:
            spec_str: Package specification string (e.g. "requests>=2.25.0[security]")

        Returns:
            Dependency: Parsed dependency information

        Raises:
            ValueError: If the specification cannot be parsed
        """
        try:
            # TODO: This is a very basic implementation and not well tested
            # our main target for now is to get the package name.
            from packaging.requirements import Requirement

            include_specifier = False

            # Handle @ operator (package@version)
            if "@" in spec_str and not spec_str.startswith("git+"):
                name = spec_str.split("@")[0]

            # Handle caret requirements (package^version)
            elif "^" in spec_str:
                name = spec_str.split("^")[0]

            # Handle tilde requirements (package~version)
            elif "~" in spec_str and "~=" not in spec_str:
                name = spec_str.split("~")[0]

            else:
                # Common PEP 440 cases
                name = spec_str
                include_specifier = True

            req = Requirement(name)

            return Dependency(
                name=req.name,
                version_constraint=str(req.specifier) if include_specifier else None,
                extras=req.extras,
                arg_index=arg_index,
                original_text=spec_str,
            )
        except Exception:
            # If spec parsing fails, just ignore for now
            return None



from pathlib import Path
import sys
from typing import TYPE_CHECKING, List, Optional, Tuple
import logging
import typer

from safety.tool.utils import PoetryPyprojectConfigurator

from .constants import MSG_SAFETY_SOURCE_ADDED, MSG_SAFETY_SOURCE_NOT_ADDED
from .parser import PoetryParser

from ..base import BaseCommand, ToolIntentionType

from ..environment_diff import EnvironmentDiffTracker, PipEnvironmentDiffTracker
from safety_schemas.models.events.types import ToolType

from safety.console import main_console as console

PO_LOCK = "safety-po.lock"

if TYPE_CHECKING:
    from ..environment_diff import EnvironmentDiffTracker

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

logger = logging.getLogger(__name__)


class PoetryCommand(BaseCommand):
    """
    Main class for hooks into poetry commands.
    """

    def get_tool_type(self) -> ToolType:
        return ToolType.POETRY

    def get_command_name(self) -> List[str]:
        return ["poetry"]

    def get_lock_path(self) -> str:
        return PO_LOCK

    def get_diff_tracker(self) -> "EnvironmentDiffTracker":
        # pip diff tracker will be enough for poetry
        return PipEnvironmentDiffTracker()

    def should_track_state(self) -> bool:
        """
        Determine if the Poetry command will change package dependencies in the virtual environment.

        Returns:
            bool: True if command will modify installed packages, False otherwise
        """
        command_str = " ".join(self._args).lower()

        package_modifying_commands = [
            "add",
            "remove",
            "install",
            "uninstall",
            "update",
            "sync",
        ]

        return any(cmd in command_str for cmd in package_modifying_commands)

    def get_package_list_command(self) -> List[str]:
        """
        Get the package list of a poetry virtual environment.

        This implementation uses pip to list packages.

        Returns:
            List[str]: Command to list packages in JSON format
        """
        return ["poetry", "run", "pip", "list", "--format=json"]

    def _has_safety_source_in_pyproject(self) -> bool:
        """
        Check if 'safety' source exists in pyproject.toml
        """
        if not Path("pyproject.toml").exists():
            return False

        try:
            # Parse the TOML file
            with open("pyproject.toml", "rb") as f:
                pyproject = tomllib.load(f)

            poetry_config = pyproject.get("tool", {}).get("poetry", {})

            sources = poetry_config.get("source", [])
            if isinstance(sources, dict):
                return "safety" in sources
            else:
                return any(source.get("name") == "safety" for source in sources)

        except (FileNotFoundError, KeyError, tomllib.TOMLDecodeError):
            return False

    def before(self, ctx: typer.Context):
        super().before(ctx)

        if self._intention and self._intention.intention_type in [
            ToolIntentionType.SYNC_PACKAGES,
            ToolIntentionType.ADD_PACKAGE,
        ]:
            if not self._has_safety_source_in_pyproject():
                org_slug = None
                try:
                    data = ctx.obj.auth.client.initialize()
                    org_slug = data.get("organization-data", {}).get("slug")
                except Exception:
                    logger.exception(
                        "Unable to pull the org slug from the initialize endpoint."
                    )

                try:
                    configurator = PoetryPyprojectConfigurator()
                    prj_slug = ctx.obj.project.id if ctx.obj.project else None
                    if configurator.configure(
                        Path("pyproject.toml"), org_slug, prj_slug
                    ):
                        console.print(
                            MSG_SAFETY_SOURCE_ADDED,
                        )
                except Exception:
                    logger.exception("Unable to configure the pyproject.toml file.")
                    console.print(
                        MSG_SAFETY_SOURCE_NOT_ADDED,
                    )

    @classmethod
    def from_args(cls, args: List[str], **kwargs):
        parser = PoetryParser()

        if intention := parser.parse(args):
            kwargs["intention"] = intention
            if intention.intention_type is ToolIntentionType.ADD_PACKAGE:
                return PoetryAddCommand(args, **kwargs)

        return PoetryGenericCommand(args, **kwargs)


class PoetryGenericCommand(PoetryCommand):
    pass


class PoetryAddCommand(PoetryCommand):
    def patch_source_option(
        self, args: List[str], new_source: str = "safety"
    ) -> Tuple[Optional[str], List[str]]:
        """
        Find --source argument and its value in a list of args, create a modified copy
        with your custom source, and return both.

        Args:
            args: List[str] - Command line arguments

        Returns:
            tuple: (source_value, modified_args, original_args)
        """
        source_value = None
        modified_args = args.copy()

        for i in range(len(args)):
            if args[i].startswith("--source="):
                # Handle --source=value format
                source_value = args[i].split("=", 1)[1]
                modified_args[i] = f"--source={new_source}"
                break
            elif args[i] == "--source" and i < len(args) - 1:
                # Handle --source value format
                source_value = args[i + 1]
                modified_args[i + 1] = new_source
                break

        return source_value, modified_args

    def before(self, ctx: typer.Context):
        super().before(ctx)

        _, modified_args = self.patch_source_option(self._args)
        self._args = modified_args



from .tool_inspector import ToolInspector
from .factory import tool_commands
from .main import configure_system, configure_alias
from .base import ToolResult


__all__ = [
    "ToolInspector",
    "tool_commands",
    "configure_system",
    "configure_alias",
    "ToolResult",
]



import logging
import re
import shutil
import subprocess
from pathlib import Path
from typing import Optional
from urllib.parse import urlsplit, urlunsplit

import typer
from rich.console import Console

from safety.tool.constants import (
    PUBLIC_REPOSITORY_URL,
    ORGANIZATION_REPOSITORY_URL,
    PROJECT_REPOSITORY_URL,
)
from safety.tool.resolver import get_unwrapped_command

from safety.console import main_console
from safety.tool.auth import index_credentials


logger = logging.getLogger(__name__)


class Pip:
    @classmethod
    def is_installed(cls) -> bool:
        """
        Checks if the PIP program is installed

        Returns:
            True if PIP is installed on system, or false otherwise
        """
        return shutil.which("pip") is not None

    @classmethod
    def configure_requirements(
        cls,
        file: Path,
        org_slug: Optional[str],
        project_id: Optional[str],
        console: Console = main_console,
    ) -> Optional[Path]:
        """
        Configures Safety index url for specified requirements file.

        Args:
            file (Path): Path to requirements.txt file.
            org_slug (str): Organization slug.
            project_id (str): Project identifier.
            console (Console): Console instance.
        """

        with open(file, "r+") as f:
            content = f.read()

            repository_url = (
                PROJECT_REPOSITORY_URL.format(org_slug, project_id)
                if project_id and org_slug
                else (
                    ORGANIZATION_REPOSITORY_URL.format(org_slug)
                    if org_slug
                    else PUBLIC_REPOSITORY_URL
                )
            )
            index_config = f"-i {repository_url}\n"
            if content.find(index_config) == -1:
                f.seek(0)
                f.write(index_config + content)

                logger.info(f"Configured {file} file")
                return file
            else:
                logger.info(f"{file} is already configured. Skipping.")

        return None

    @classmethod
    def configure_system(
        cls, org_slug: Optional[str], console: Console = main_console
    ) -> Optional[Path]:
        """
        Configures PIP system to use to Safety index url.
        """
        try:
            repository_url = (
                ORGANIZATION_REPOSITORY_URL.format(org_slug)
                if org_slug
                else PUBLIC_REPOSITORY_URL
            )
            result = subprocess.run(
                [
                    get_unwrapped_command(name="pip"),
                    "config",
                    "set",
                    "global.index-url",
                    repository_url,
                ],
                capture_output=True,
            )

            if result.returncode != 0:
                logger.error(
                    f"Failed to configure PIP global settings: {result.stderr.decode('utf-8')}"
                )
                return None

            output = result.stdout.decode("utf-8")
            match = re.search(r"Writing to (.+)", output)

            if match:
                config_file_path = match.group(1).strip()
                return Path(config_file_path)

            logger.error("Failed to match the config file path written by pip.")
            return Path()
        except Exception:
            logger.exception("Failed to configure PIP global settings.")

        return None

    @classmethod
    def reset_system(cls, console: Console = main_console):
        # TODO: Move this logic and implement it in a more robust way
        try:
            subprocess.run(
                [
                    get_unwrapped_command(name="pip"),
                    "config",
                    "unset",
                    "global.index-url",
                ],
                capture_output=True,
            )
        except Exception:
            console.print("Failed to reset PIP global settings.")

    @classmethod
    def default_index_url(cls) -> str:
        return "https://pypi.org/simple/"

    @classmethod
    def build_index_url(cls, ctx: typer.Context, index_url: Optional[str]) -> str:
        if index_url is None:
            index_url = PUBLIC_REPOSITORY_URL

        url = urlsplit(index_url)

        encoded_auth = index_credentials(ctx)
        netloc = f"user:{encoded_auth}@{url.netloc}"

        if type(url.netloc) is bytes:
            url = url._replace(netloc=netloc.encode("utf-8"))
        elif type(url.netloc) is str:
            url = url._replace(netloc=netloc)

        return urlunsplit(url)



from sys import platform
import subprocess


def get_unwrapped_command(name: str) -> str:
    """
    Find the true executable for a command, skipping wrappers/aliases/.bat files.

    Args:
        command: The command to resolve (e.g. 'pip', 'python')

    Returns:
        Path to the actual executable
    """
    if platform in ["win32"]:
        lookup_term = f"{name}.exe"
        where_result = subprocess.run(
            ["where.exe", lookup_term], capture_output=True, text=True
        )
        if where_result.returncode == 0:
            for path in where_result.stdout.splitlines():
                if not path.lower().endswith(f"{name}.bat"):
                    return path

    return name



import logging
from typing import Dict, Optional, Tuple, Any, Callable

import click

from .main import (
    get_auth_info,
    get_host_config,
    get_organization,
    get_proxy_config,
    get_redirect_url,
    get_token_data,
    save_auth_config,
    get_token,
    clean_session,
)
from authlib.common.security import generate_token
from safety.auth.constants import CLIENT_ID, OPENID_CONFIG_URL

from safety.auth.models import Organization, Auth
from safety.auth.utils import (
    S3PresignedAdapter,
    SafetyAuthSession,
    get_keys,
    is_email_verified,
)
from safety.constants import REQUEST_TIMEOUT
from safety.scan.constants import (
    CLI_KEY_HELP,
    CLI_PROXY_HOST_HELP,
    CLI_PROXY_PORT_HELP,
    CLI_PROXY_PROTOCOL_HELP,
    CLI_STAGE_HELP,
)
from safety.util import DependentOption, SafetyContext, get_proxy_dict
from safety.models import SafetyCLI
from safety_schemas.models import Stage

LOG = logging.getLogger(__name__)


def build_client_session(
    api_key: Optional[str] = None,
    proxies: Optional[Dict[str, str]] = None,
    headers: Optional[Dict[str, str]] = None,
) -> Tuple[SafetyAuthSession, Dict[str, Any]]:
    """
    Builds and configures the client session for authentication.

    Args:
        api_key (Optional[str]): The API key for authentication.
        proxies (Optional[Dict[str, str]]): Proxy configuration.
        headers (Optional[Dict[str, str]]): Additional headers.

    Returns:
        Tuple[SafetyAuthSession, Dict[str, Any]]: The configured client session and OpenID configuration.
    """

    kwargs = {}
    target_proxies = proxies

    # Global proxy defined in the config.ini
    proxy_config, proxy_timeout, proxy_required = get_proxy_config()

    if not proxies:
        target_proxies = proxy_config

    def update_token(tokens, **kwargs):
        save_auth_config(
            access_token=tokens["access_token"],
            id_token=tokens["id_token"],
            refresh_token=tokens["refresh_token"],
        )
        load_auth_session(click_ctx=click.get_current_context(silent=True))  # type: ignore

    client_session = SafetyAuthSession(
        client_id=CLIENT_ID,
        code_challenge_method="S256",
        redirect_uri=get_redirect_url(),
        update_token=update_token,
        scope="openid email profile offline_access",
        **kwargs,
    )

    client_session.mount("https://pyup.io/static-s3/", S3PresignedAdapter())

    client_session.proxy_required = proxy_required
    client_session.proxy_timeout = proxy_timeout
    client_session.proxies = target_proxies  # type: ignore
    client_session.headers = {
        "Accept": "application/json",
        "Content-Type": "application/json",
    }

    try:
        openid_config = client_session.get(
            url=OPENID_CONFIG_URL, timeout=REQUEST_TIMEOUT
        ).json()
    except Exception as e:
        LOG.debug("Unable to load the openID config: %s", e)
        openid_config = {}

    client_session.metadata["token_endpoint"] = openid_config.get(
        "token_endpoint", None
    )

    if api_key:
        client_session.api_key = api_key  # type: ignore
        client_session.headers["X-Api-Key"] = api_key

    if headers:
        client_session.headers.update(headers)

    return client_session, openid_config


def load_auth_session(click_ctx: click.Context) -> None:
    """
    Loads the authentication session from the context.

    Args:
        click_ctx (click.Context): The Click context object.
    """
    if not click_ctx:
        LOG.warning("Click context is needed to be able to load the Auth data.")
        return

    client = click_ctx.obj.auth.client
    keys = click_ctx.obj.auth.keys

    access_token: str = get_token(name="access_token")  # type: ignore
    refresh_token: str = get_token(name="refresh_token")  # type: ignore
    id_token: str = get_token(name="id_token")  # type: ignore

    if access_token and keys:
        try:
            token = get_token_data(access_token, keys, silent_if_expired=True)
            client.token = {
                "access_token": access_token,
                "refresh_token": refresh_token,
                "id_token": id_token,
                "token_type": "bearer",
                "expires_at": token.get("exp", None),  # type: ignore
            }
        except Exception as e:
            print(e)
            clean_session(client)


def proxy_options(func: Callable) -> Callable:
    """
    Decorator that defines proxy options for Click commands.

    Options defined per command, this will override the proxy settings defined in the
    config.ini file.

    Args:
        func (Callable): The Click command function.

    Returns:
        Callable: The wrapped Click command function with proxy options.
    """
    func = click.option(
        "--proxy-protocol",
        type=click.Choice(["http", "https"]),
        default="https",
        cls=DependentOption,
        required_options=["proxy_host"],
        help=CLI_PROXY_PROTOCOL_HELP,
    )(func)
    func = click.option(
        "--proxy-port",
        multiple=False,
        type=int,
        default=80,
        cls=DependentOption,
        required_options=["proxy_host"],
        help=CLI_PROXY_PORT_HELP,
    )(func)
    func = click.option(
        "--proxy-host", multiple=False, type=str, default=None, help=CLI_PROXY_HOST_HELP
    )(func)

    return func


def auth_options(stage: bool = True) -> Callable:
    """
    Decorator that defines authentication options for Click commands.

    Args:
        stage (bool): Whether to include the stage option.

    Returns:
        Callable: The decorator function.
    """

    def decorator(func: Callable) -> Callable:
        func = click.option(
            "--key", default=None, envvar="SAFETY_API_KEY", help=CLI_KEY_HELP
        )(func)

        if stage:
            func = click.option(
                "--stage", default=None, envvar="SAFETY_STAGE", help=CLI_STAGE_HELP
            )(func)

        return func

    return decorator


def inject_session(
    ctx: click.Context,
    proxy_protocol: Optional[str] = None,
    proxy_host: Optional[str] = None,
    proxy_port: Optional[str] = None,
    key: Optional[str] = None,
    stage: Optional[Stage] = None,
    invoked_command: str = "",
) -> Any:
    org: Optional[Organization] = get_organization()

    if not stage:
        host_stage = get_host_config(key_name="stage")
        stage = host_stage if host_stage else Stage.development

    proxy_config: Optional[Dict[str, str]] = get_proxy_dict(
        proxy_protocol,  # type: ignore
        proxy_host,  # type: ignore
        proxy_port,  # type: ignore
    )

    client_session, openid_config = build_client_session(
        api_key=key, proxies=proxy_config
    )
    keys = get_keys(client_session, openid_config)

    auth = Auth(
        stage=stage,
        keys=keys,
        org=org,
        client_id=CLIENT_ID,  # type: ignore
        client=client_session,
        code_verifier=generate_token(48),
    )

    if not ctx.obj:
        ctx.obj = SafetyCLI()

    ctx.obj.auth = auth

    load_auth_session(ctx)

    info = get_auth_info(ctx)

    if info:
        ctx.obj.auth.name = info.get("name")
        ctx.obj.auth.email = info.get("email")
        ctx.obj.auth.email_verified = is_email_verified(info)  # type: ignore
        SafetyContext().account = info["email"]
    else:
        SafetyContext().account = ""

    @ctx.call_on_close
    def clean_up_on_close():
        LOG.debug("Closing requests session.")
        ctx.obj.auth.client.close()

        if ctx.obj.event_bus:
            from safety.events.utils import (
                create_internal_event,
                InternalEventType,
                InternalPayload,
            )

            payload = InternalPayload(ctx=ctx)

            flush_event = create_internal_event(
                event_type=InternalEventType.FLUSH_SECURITY_TRACES, payload=payload
            )
            close_event = create_internal_event(
                event_type=InternalEventType.CLOSE_RESOURCES, payload=payload
            )

            flush_future = ctx.obj.event_bus.emit(flush_event)
            close_future = ctx.obj.event_bus.emit(close_event)

            # Wait for both events to be processed
            if flush_future and close_future:
                try:
                    flush_future.result()
                    close_future.result()
                except Exception as e:
                    LOG.warning(f"Error waiting for events to process: {e}")

            ctx.obj.event_bus.stop()



from safety.models import Package, Vulnerability, SafetyRequirement
from packaging.specifiers import SpecifierSet

# JSON report loaded
VALID_REPORT = {'report_meta': {'scan_target': 'environment', 'scanned': ['/usr/local/lib/python3.9/site-packages'],
                                'api_key': True, 'packages_found': 50, 'local_database_path_used': None,
                                'timestamp': '2022-04-30 08:16:30',
                                'safety_version': '2.0b1'},
                'scanned_packages': {'safety': {'name': 'safety', 'version': '2.0b1',
                                                'requirements': [{
                                                    "raw": "safety==2.0b1",
                                                    "extras": [],
                                                    "marker": None,
                                                    "name": "safety",
                                                    "specifier": "==2.0b1",
                                                    "url": None
                                                }]
                                                },
                                     'click': {'name': 'click', 'version': '8.0.4',
                                               'requirements': [{
                                                   "raw": "click==8.0.4",
                                                   "extras": [],
                                                   "marker": None,
                                                   "name": "click",
                                                   "specifier": "==8.0.4",
                                                   "url": None
                                               }]},
                                     'insecure-package': {'name': 'insecure-package', 'version': '0.1.0',
                                                          'requirements': [
                                                              {
                                                                  "raw": "insecure-package==0.1.0",
                                                                  "extras": [],
                                                                  "marker": None,
                                                                  "name": "insecure-package",
                                                                  "specifier": "==0.1.0",
                                                                  "url": None
                                                              }]}},
                'affected_packages': {'insecure-package': {'name': 'insecure-package', 'version': '0.1.0',
                                                           'requirements':
                                                               [
                                                                   {
                                                                       "raw": "insecure-package==0.1.0",
                                                                       "extras": [],
                                                                       "marker": None,
                                                                       "name": "insecure-package",
                                                                       "specifier": "==0.1.0",
                                                                       "url": None
                                                                   }
                                                               ],
                                                           'found': '/usr/local/lib/python3.9/site-packages',
                                                           'insecure_versions': ['0.1.0'], 'secure_versions': [],
                                                           'latest_version_without_known_vulnerabilities': '',
                                                           'latest_version': '0.1.0',
                                                           'more_info_url': 'https://pyup.io/packages/pypi/insecure-package/'}},
                'announcements': [], 'vulnerabilities': [
        {'vulnerability_id': '25853', 'package_name': 'insecure-package', 'ignored': False, 'ignored_reason': None,
         'ignored_expires': None, 'vulnerable_spec': ['<0.2.0'],
         'all_vulnerable_specs': ['<0.2.0'], 'analyzed_version': '0.1.0', 'analyzed_requirement': {
                                                                  "raw": "insecure-package==0.1.0",
                                                                  "extras": [],
                                                                  "marker": None,
                                                                  "name": "insecure-package",
                                                                  "specifier": "==0.1.0",
                                                                  "url": None
                                                              },
         'advisory': 'This is an insecure package with lots of exploitable security vulnerabilities.',
         'is_transitive': False, 'published_date': '2021-Apr-14',
         'fixed_versions': [''], 'closest_versions_without_known_vulnerabilities': [],
         'resources': ['https://pypi.org/project/insecure-package'],
         'CVE': None,
         'severity': None, 'affected_versions': ['0.1.0'],
         'more_info_url': 'https://pyup.io/vulnerabilities/PVE-2021-25853/25853/'}],
                'ignored_vulnerabilities': [],
                'remediations': {
                    'insecure-package': {
                        'requirements': {
                            '==0.1.0': {
                                'version': '0.1.0',
                                'requirement': {
                                    "raw": "insecure-package==0.1.0",
                                    "extras": [],
                                    "marker": None,
                                    "name": "insecure-package",
                                    "specifier": "==0.1.0",
                                    "url": None
                                },
                                'vulnerabilities_found': 1, 'recommended_version': None,
                                'other_recommended_versions': [],
                                'more_info_url': 'https://pyup.io/packages/pypi/insecure-package/'
                            },
                        }}}}


# Python objects
SCANNED_PACKAGES = {
    'safety': Package(name='safety', version='2.0b1', requirements=[SafetyRequirement('safety==2.0b1')]),
    'click': Package(name='click', version='8.0.4', requirements=[SafetyRequirement('click==8.0.4')]),
    'insecure-package': Package(name='insecure-package', version='0.1.0',
                                requirements=[SafetyRequirement('insecure-package==0.1.0')]),
}

VULNS = [Vulnerability(vulnerability_id='25853', package_name='insecure-package',
                       pkg=Package(name='insecure-package', version='0.1.0',
                                   requirements=[SafetyRequirement('insecure-package==0.1.0')],
                                   found=None,
                                   insecure_versions=[], secure_versions=[],
                                   latest_version_without_known_vulnerabilities=None,
                                   latest_version=None,
                                   more_info_url=None),
                       ignored=False, ignored_reason=None, ignored_expires=None, vulnerable_spec=['<0.2.0'],
                       all_vulnerable_specs=['<0.2.0'], analyzed_version='0.1.0',
                       analyzed_requirement=SafetyRequirement('insecure-package==0.1.0'),
                       advisory='This is an insecure package with lots of exploitable security vulnerabilities.',
                       is_transitive=False, published_date='2021-Apr-14', fixed_versions=[''],
                       closest_versions_without_known_vulnerabilities=[],
                       resources=['https://pypi.org/project/insecure-package'],
                       CVE=None, severity=None, affected_versions=['0.1.0'],
                       more_info_url='https://pyup.io/vulnerabilities/PVE-2021-25853/25853/')]


REMEDIATIONS = {
    'insecure-package': {
        '==0.1.0': {
            'version': '0.1.0',
            'requirement': SafetyRequirement('insecure-package==0.1.0'),
            'vulnerabilities_found': 1, 'recommended_version': None,
            'other_recommended_versions': [],
            'more_info_url': 'https://pyup.io/packages/pypi/insecure-package/'
        },
    }
}
